#!/usr/bin/python
# Copyright (c) 2016-2021 Julien Peloton, Giulio Fabbian.
#
# This file is part of s4cmb
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Module to simulate pointing of the focal plane detectors.

Reminder mostly for myself:
Az/El specify the direction of a point on the celestial sphere in the
horizontal coordinate system (a spherical coordinate system).
RA/Dec specify the direction of a point on the celestial sphere in the
equatorial coordinate system.

Author: Julien Peloton, peloton@lal.in2p3.fr
        Giulio Fabbian, g.fabbian@sussex.ac.uk
"""
from __future__ import division, absolute_import, print_function

import os
import numpy as np
from numpy import cos
from numpy import sin
from numpy import tan
from pyslalib import slalib
from astropy.utils import iers
from astropy.time import Time
import warnings
from s4cmb.detector_pointing_f import detector_pointing_f

sec2deg = 360.0 / 86400.0
d2r = np.pi / 180.0
ASTROMETRIC_GEOCENTRIC = 0
APPARENT_GEOCENTRIC = 1
APPARENT_TOPOCENTRIC = 2


def get_ut1utc(ut1utc_fn, mjd):
    """Return the time correction to UTC.

    Those are computed from 01 01 2012 (MJD=55927) to the latest MJD available
    in the s4cmb/data/ut1utc.ephem file.
    Unfortunately the IERS Bulletin A database used to compute those is often
    down/flaky (http://maia.usno.navy.mil/cgi-bin/search.cgi). For dates outside
    the range included in the ut1utc.ephem file, the 01 01 2012 correction is
    applied for earlier dates and the correction of the last available date in
    the file is applied for future dates. Consider updating the content of your
    ut1utc file using the dedicated routine update_ut1utc in s4cmb (see below,
    uses astropy and requires internet access).

    Parameters
    ----------
    ut1utc_fn : string
        Filename where ut1utc are stored.
    mjd : float
        Date (in MJD) to correct for.

    Returns
    ----------
    ut1utc : float
        Contain the time correction to apply to MJD values.

    Examples
    ----------
    >>> fn = 's4cmb/data/ut1utc.ephem'
    >>> print(round(get_ut1utc(fn, mjd=56293), 3))
    0.277
    """
    warn_me = True
    umjds, ut1utcs = np.loadtxt(ut1utc_fn, usecols=(1, 2)).T
    if mjd < umjds[0]:
        mjd_tmp = umjds[0]
    elif mjd > umjds[-1]:
        mjd_tmp = umjds[-1]
    else:
        mjd_tmp = mjd
        warn_me = False
    if warn_me:
        warnings.warn(
            """
            MJD outside ut1utc file date range {:.6f} -- {:.6f}.
            Default corrections applied. Consider updating the ut1utc file.
            """.format(umjds[0], umjds[-1])
        )
    uindex = np.searchsorted(umjds, mjd_tmp)
    ut1utc = ut1utcs[uindex]
    return ut1utc


class Pointing:
    """ Class to handle detector pointing """

    def __init__(
        self,
        az_enc,
        el_enc,
        time,
        value_params,
        allowed_params="ia ie ca an aw",
        ra_src=0.0,
        dec_src=0.0,
        lat=-22.958,
        ut1utc_fn="s4cmb/data/ut1utc.ephem",
    ):
        """
        Apply pointing model with parameters `value_params` and
        names `allowed_params` to encoder az,el. Order of terms is
        `value_params` is same as order of terms in `allowed_params`.

        Full list of parameters (Thx Fred!):
            an:  azimuth axis tilt north of vertical
            aw:  azimuth axis tilt west of vertical
            an2:  potato chip
            aw2:  potato chip
            npae:  not parallel azimuth/elevation
            ca:  angle of beam to boresight in azimuth
            ia:  azimuth encoder zero
            ie:  elevation encoder zero + angle of beam
                to boresight in elevation
            tf:  cosine flexure
            tfs:  sine flexure
            ref:  refraction
            dt:  timing error in seconds (requires lat argument)
            elt:  time from start elevation correction
            ta1: linear order structural thermal warping in azimuth
            te1: linear order structural thermal warping in elevation
            sa,sa2: solar radiation structural warping in azimuth
            se,se2: solar radiation structural warping in elevation

        Parameters
        ----------
        az_enc : 1d array
            Encoder azimuth in radians.
        el_enc : 1d array
            Encoder elevation in radians.
        time : 1d array
            Encoder time (UTC) in mjd
        value_params : 1d array
            Value of the pointing model parameters (see instrument.py).
            In degrees (see below for full description)
        allowed_params : list of string, optional
            Name of the pointing model parameters used in `value_params`.
        ra_src : float, optional
            RA of the reference (keep it to zero).
        dec_src : float, optional
            Dec of the reference (keep it to zero).
        lat : float, optional
            Latitude of the telescope, in degree.
        ut1utc_fn : string, optional
            File containing time correction to UTC.
            The \Delta{UT} (UT1-UTC) is tabulated in IERS circulars
            and elsewhere. It increases by exactly one second at the end
            of each UTC leap second, introduced in order to
            keep \Delta{UT} within \pm 0s.9.
            The 'sidereal \Delta{UT}' which forms part of AOPRMS(13)
            is the same quantity, but converted from solar to sidereal
            seconds and expressed in radians.
            WTF?

        Examples
        ----------
        See hardware.py for more information on the pointing model.
        >>> allowed_params, value_params, az_enc, el_enc, time = \
            load_fake_pointing()
        >>> pointing = Pointing(az_enc, el_enc, time, value_params,
        ...     allowed_params, lat=-22.)
        >>> print(round(az_enc[2], 2), round(pointing.az[2], 2))
        0.13 0.12
        """
        self.az_enc = az_enc
        self.el_enc = el_enc
        self.time = time
        self.value_params = value_params
        self.allowed_params = allowed_params
        self.lat = lat * d2r
        self.ut1utc_fn = ut1utc_fn
        self.ra_src = ra_src
        self.dec_src = dec_src

        self.ut1utc = get_ut1utc(self.ut1utc_fn, self.time[0])

        # Initialise the object
        self.az, self.el = self.apply_pointing_model()
        self.azel2radec()

    def apply_pointing_model(self):
        """
        Apply pointing corrections specified by the pointing model.

        Returns
        ----------
        az : 1d array
            The corrected azimuth in arcminutes.
        el : 1d array
            The corrected elevation in arcminutes.
        """
        assert len(self.value_params) == len(
            self.allowed_params.split()
        ), AssertionError(
            """Vector containing parameters
            (value_params) has to have the same
            length than the vector containing names
            (allowed_params)."""
        )

        # Here are many parameters defining a pointing model.
        # Of course, we do not use all of them. They are zero by default,
        # and only those specified by the user will be used.
        params = {
            p: 0.0
            for p in [
                "an",
                "aw",
                "an2",
                "aw2",
                "an4",
                "aw4",
                "npae",
                "ca",
                "ia",
                "ie",
                "tf",
                "tfs",
                "ref",
                "dt",
                "elt",
                "ta1",
                "te1",
                "sa",
                "se",
                "sa2",
                "se2",
                "sta",
                "ste",
                "sta2",
                "ste2",
            ]
        }

        for param in params:
            if param in self.allowed_params.split():
                index = self.allowed_params.split().index(param)
                params[param] = self.value_params[index]

        params["dt"] *= sec2deg

        # Azimuth
        azd = -params["an"] * sin(self.az_enc) * sin(self.el_enc)
        azd -= params["aw"] * cos(self.az_enc) * sin(self.el_enc)

        azd -= -params["an2"] * sin(2 * self.az_enc) * sin(self.el_enc)
        azd -= params["aw2"] * cos(2 * self.az_enc) * sin(self.el_enc)

        azd -= -params["an4"] * sin(4 * self.az_enc) * sin(self.el_enc)
        azd -= params["aw4"] * cos(4 * self.az_enc) * sin(self.el_enc)

        azd += params["npae"] * sin(self.el_enc)
        azd -= params["ca"]
        azd += params["ia"] * cos(self.el_enc)

        azd += params["dt"] * (
            -sin(self.lat) + cos(self.az_enc) * cos(self.lat) * tan(self.el_enc)
        )

        # Elevation
        eld = params["an"] * cos(self.az_enc)
        eld -= params["aw"] * sin(self.az_enc)
        eld -= params["an2"] * cos(2 * self.az_enc)
        eld -= params["aw2"] * sin(2 * self.az_enc)
        eld -= params["an4"] * cos(4 * self.az_enc)
        eld -= params["aw4"] * sin(4 * self.az_enc)

        eld -= params["ie"]
        eld += params["tf"] * cos(self.el_enc)
        eld += params["tfs"] * sin(self.el_enc)
        eld -= params["ref"] / tan(self.el_enc)

        eld += -params["dt"] * cos(self.lat) * sin(self.az_enc)

        eld += params["elt"] * (self.time - np.min(self.time))

        # Convert back in radian and apply to the encoder values.
        azd *= np.pi / (180.0 * 60.0)
        eld *= np.pi / (180.0 * 60.0)

        azd /= np.cos(self.el_enc)

        az = self.az_enc - azd
        el = self.el_enc - eld

        return az, el

    def azel2radec(self):
        """
        Given Az/El, time, and time correction returns RA/Dec, parallactic
        angles, and quaternions (used later to rotate detector coordinates).

        Examples
        ----------
        Go from az/el -> ra/dec
        >>> allowed_params, value_params, az_enc, el_enc, time = \
            load_fake_pointing()
        >>> pointing = Pointing(az_enc, el_enc, time, value_params,
        ...     allowed_params, lat=-22.)
        >>> print(round(pointing.ra[2], 2), round(pointing.dec[2], 2))
        0.7 0.66
        """
        self.ra, self.dec, self.pa = self.azel2radecpa()
        v_ra = self.ra
        v_dec = self.dec
        v_pa = self.pa
        v_ra_src = self.ra_src
        v_dec_src = self.dec_src

        self.meanpa = np.median(v_pa)

        self.quaternion = Quaternion(v_ra, v_dec, v_pa, v_ra_src, v_dec_src)

        q = self.quaternion.offset_radecpa_makequat()

        assert q.shape == (self.az.size, 4), AssertionError(
            "Wrong size for the quaternions!"
        )

        self.q = q

    def azel2radecpa(self):
        """
        Given Az/El, time, and time correction returns RA/Dec and parallactic
        angles.

        Examples
        ----------
        Go from az/el -> ra/dec/pa
        >>> allowed_params, value_params, az_enc, el_enc, time = \
            load_fake_pointing()
        >>> pointing = Pointing(az_enc, el_enc, time, value_params,
        ...     allowed_params, lat=-22.)
        >>> ra, dec, pa = pointing.azel2radecpa()
        >>> print(round(ra[0], 2), round(dec[0], 2), round(pa[0], 2))
        0.56 0.67 3.13
        """
        # TODO pass lon, lat, etc from the ScanningStrategy module!
        converter = Azel2Radec(self.time[0], self.ut1utc)
        vconv = np.vectorize(converter.azel2radecpa)
        ra, dec, pa = vconv(self.time, self.az, self.el)
        return ra, dec, pa

    def radec2azel(self):
        """
        Given RA/Dec, time, and time correction returns Az/El.

        Examples
        ----------
        Make sure that we go back to our feet az/el -> ra/dec -> az/el
        >>> allowed_params, value_params, az_enc, el_enc, time = \
            load_fake_pointing()
        >>> pointing = Pointing(az_enc, el_enc, time, value_params,
        ...     allowed_params, lat=-22.)
        >>> az, el = pointing.radec2azel()
        >>> assert np.all(np.round(az[2:4],2) == np.round(pointing.az[2:4],2))
        >>> assert np.all(np.round(el[2:4],2) == np.round(pointing.el[2:4],2))
        """
        # TODO pass lon, lat, etc from the ScanningStrategy module!
        converter = Azel2Radec(self.time[0], self.ut1utc)
        vconv = np.vectorize(converter.radec2azel)
        az, el = vconv(self.time, self.ra, self.dec)
        return az, el

    def offset_detector(self, azd, eld):
        """
        To compute RA/Dec of each detector from az/el, it is much
        faster to use the quaternions. This routine does it for you.

        Parameters
        ----------
        azd : 1d array
            The azimuth array for the observation in radian.
        els : 1d array
            The elevation array for the observation in radian.

        Returns
        ----------
        ra : 1d array
            Right ascension in radian.
        dec : 1d array
            Declination in radian.
        pa : 1d array
            Parallactic angle in radian.
        """
        ra, dec, pa = self.quaternion.offset_radecpa_applyquat(self.q, -azd, -eld)
        return ra, dec, pa


class Azel2Radec(object):
    """ Class to handle az/el <-> ra/dec conversion """

    def __init__(
        self,
        mjd,
        ut1utc,
        lon=-67.786,
        lat=-22.958,
        height=5200.0,
        pressure=533.29,
        temp=273.15,
        humidity=0.1,
        epequi=2000.0,
    ):
        """
        This class is mostly a wrapper around slalib.
        The default parameters correspond to the Polarbear observation site.

        Parameters
        ----------
        mjd : float
            Date in MJD.
        ut1utc : float
            Time correction (UT1 - UTC) for the date.
        lon : float, optional
            Longitute of the site of observation in degree.
        lat : float, optional
            Latitude of the site of observation in degree.
        height : float, optional
            Altitude of the site of observation.
        pressure : float, optional
            Local atmospheric pressure at the site of observation in millibar.
        temp : float, optional
            Local ambient temperature at the site of observation in K.
        humidity : float, optional
            Local relative humidity at the site of observation
            (value between 0 and 1).
        epequi : float, optional
            Epoch of mean equinox to be used (Julian).

        """
        self.lon = lon * d2r
        self.lat = lat * d2r
        self.height = height
        self.pressure = pressure
        self.temp = temp
        self.humidity = humidity
        self.mjd = mjd

        self.epequi = epequi
        self.ut1utc = ut1utc

        self.updateaoprms(mjd)

    def updateaoprms(self, mjd, lapserate=0.0065, xpm=0.0, ypm=0.0):
        """
        Pre-compute the set of apparent to observed place parameters.
        See http://star-www.rl.ac.uk/docs/sun67.htx/sun67ss8.html

        Parameters
        ----------
        lapserate : float
            ta mere
        xpm : float
            polar motion coordinate x (radians)
        ypm : float
            polar motion coordinate y (radians)

        """
        wavelength = (299792458.0 / 150.0e9) * 1e6
        self.aoprms = slalib.sla_aoppa(
            mjd,
            self.ut1utc,
            self.lon,
            self.lat,
            self.height,
            xpm,
            ypm,
            self.temp,
            self.pressure,
            self.humidity,
            wavelength,
            lapserate,
        )

    def azel2radecpa(self, mjd, az, el):
        """
        Given Az/El and time returns RA/Dec and parallactic angle.
        This routine does not return a precisely correct parallactic angle.

        Parameters
        ----------
        mjd : float
            Date in MJD.
        az : float
            Azimuth in radian.
        el : float
            Elevation in radian.

        Returns
        ----------
        ra : float
            Right ascension in radian.
        dec : float
            Declination in radian.
        pa : float
            Parallactic angle in radian.
        """
        zd = np.pi / 2 - el
        amprms = slalib.sla_mappa(self.epequi, mjd)
        self.aoprms = slalib.sla_aoppat(mjd, self.aoprms)

        ra_app1, dec_app1 = slalib.sla_oapqk("a", az, zd + 1e-8, self.aoprms)
        ra1, dec1 = slalib.sla_ampqk(ra_app1, dec_app1, amprms)
        ra_app2, dec_app2 = slalib.sla_oapqk("a", az, zd - 1e-8, self.aoprms)
        ra2, dec2 = slalib.sla_ampqk(ra_app2, dec_app2, amprms)
        pa = slalib.sla_dbear(ra1, dec1, ra2, dec2)
        ra = 0.5 * (ra1 + ra2)
        dec = 0.5 * (dec1 + dec2)

        return ra, dec, pa

    def radec2azel(self, mjd, ra, dec):
        """
        Given RA/Dec and time returns Az/El.

        Parameters
        ----------
        mjd : float
            Date in MJD.
        ra : float
            Right ascension in radian.
        dec : float
            Declination in radian.

        Returns
        ----------
        az : float
            Azimuth in radian.
        el : float
            Elevation in radian.
        """
        amprms = slalib.sla_mappa(self.epequi, mjd)
        self.aoprms = slalib.sla_aoppat(mjd, self.aoprms)
        ra_app, dec_app = slalib.sla_mapqkz(ra, dec, amprms)
        az, zd, a, b, c = slalib.sla_aopqk(ra_app, dec_app, self.aoprms)
        el = np.pi / 2 - zd
        return az, el


class Quaternion:
    """ Class to handle quaternions """

    def __init__(self, ra, dec, pa, v_ra_src, v_dec_src):
        """
        Once you have RA/Dec coordinates of the reference detector, one
        can efficiently use quaternions to compute each detector pointing in
        the focal plane. This is rather faster than the traditional method,
        but a bit more abstract and less intuitive unfortunately...

        Parameters
        ----------
        ra : 1d array
            Right ascension in radian.
        dec : 1d array
            Declination in radian.
        pa : 1d array
            Parallactic angle in radian.
        v_ra_src : float
            RA of the source (center of the patch for example).
        v_dec_src : float
            Dec of the source (center of the patch for example).

        """
        self.ra = ra
        self.dec = dec
        self.pa = pa
        self.v_ra_src = v_ra_src
        self.v_dec_src = v_dec_src

    def offset_radecpa_makequat(self):
        """
        Construct the quaternions from initial RA/Dec coordinates.

        Returns
        ----------
        q : array
            Quaternions array.
        """
        qra = euler_quatz(self.ra)

        qdec = euler_quaty(-self.dec)
        qpa = euler_quatx(-self.pa)

        qracen = euler_quatz(-self.v_ra_src)
        qdeccen = euler_quaty(self.v_dec_src)

        q = mult(qdec, qpa)
        q = mult(qra, q)
        q = mult(qracen, q)
        q = mult(qdeccen, q)

        return q

    def offset_radecpa_applyquat(self, q, azd, eld):
        """
        Apply pre-computed quaternions to obtain
        desired RA/Dec and parallactic angle from az/el of the detector.

        Parameters
        ----------
        q : array
            Quaternions array.
        azd : 1d array
            Azimuth of the detector.
        eld : 1d array
            Elevation of the detector.

        Returns
        ----------
        psi : 1d array
            Can be identified to RA (if input -azd, -eld)
        theta : 1d array
            Can be identified to -Dec (if input -azd, -eld). Note that -theta
            is returned (so Dec directly).
        phi : 1d array
            Can be identified to -PA (if input -azd, -eld). Note that -phi
            is returned (so PA directly).
        """
        assert len(q.shape) == 2, AssertionError("Wrong quaternion size!")
        assert q.shape[1] == 4, AssertionError("Wrong quaternion size!")
        qazd = euler_quatz(-azd)
        qeld = euler_quaty(-eld)

        qpix = mult(qazd, qeld)[0]

        # Inlining this is a 30x speed up
        seq = mult_fortran(q, qpix)

        assert seq.shape[1] == 4, AssertionError("Wrong size!")

        # n = seq.shape[0]
        phi, theta, psi = quat_to_radecpa_fortran(seq)

        return psi, -theta, -phi


def radec2thetaphi(ra, dec):
    """
    Correspondance between RA/Dec and theta/phi coordinate systems.

    Parameters
    ----------
    ra : float or 1d array
        RA angle in radian.
    dec : float or 1d array
        Dec angle in radian.

    Returns
    ----------
    theta : float or 1d array
        Theta angle in radian
    phi : float or 1d array
        Phi angle in radian

    Examples
    ----------
    >>> theta, phi = radec2thetaphi(-np.pi/2., 0.)
    >>> print(round(theta, 2), round(phi, 2))
    1.57 -1.57
    """
    theta = np.pi / 2 - dec
    phi = ra
    return theta, phi


def mult(p, q):
    """
    Multiply arrays of quaternions,
    see: http://en.wikipedia.org/wiki/\
        Quaternions#Quaternions_and_the_geometry_of_R3

    Parameters
    ----------
    p : ndarray
        Array of quaternions of size (np, 4)
    q : ndarray
        Array of quaternions of size (nq, 4)

    Returns
    ----------
    pq : ndarray
        Array of size (np, 4)

    Examples
    ----------
    >>> mult(np.array([[3., 4., 5., 2.],
    ...     [2., 2., 2., 2.]]), np.array([1., 2., 3., 4.]))
    ... #doctest +NORMALIZE_WHITESPACE
    array([[ 16.,  16.,  28., -18.],
           [ 12.,   8.,  16.,  -4.]])
    """
    if p.ndim == 1 and q.ndim > 1:
        p = np.tile(p, (q.shape[0], 1))
    if q.ndim == 1 and p.ndim > 1:
        q = np.tile(q, (p.shape[0], 1))
    if q.ndim == 1 and p.ndim == 1:
        p = p.reshape((1, 4))
        q = q.reshape((1, 4))

    ps = p[:, 3]
    qs = q[:, 3]
    pv = p[:, :3]
    qv = q[:, :3]

    pq = np.empty_like(p)
    pq[:, 3] = ps * qs
    pq[:, 3] -= arraylist_dot(pv, qv).flatten()

    pq[:, :3] = ps[:, np.newaxis] * qv + pv * qs[:, np.newaxis] + np.cross(pv, qv)

    return pq


def mult_fortran(p, q):
    """
    Inline version for when p is an array of quaternions
    and q is a single quaternion. Big speed-up.

    Parameters
    ----------
    p : ndarray
        Array of quaternions of size (np, 4)
    q : ndarray
        Array of quaternions of size (1, 4)

    Returns
    ----------
    pq : ndarray
        Array of size (np, 4)

    Examples
    ----------
    >>> mult_fortran(np.array([[3., 4., 5., 2.],
    ...     [2., 2., 2., 2.]]), np.array([1., 2., 3., 4.]))
    ... #doctest +NORMALIZE_WHITESPACE
    array([[ 16.,  16.,  28., -18.],
           [ 12.,   8.,  16.,  -4.]])
    """

    assert p.ndim == 2, AssertionError("Wrong size!")
    assert p.shape[1] == 4, AssertionError("Wrong size!")
    assert q.ndim == 1, AssertionError("Wrong size!")
    assert q.size == 4, AssertionError("Wrong size!")
    pq = np.zeros_like(p)
    n = p.shape[0]

    shape = pq.shape
    pq = pq.flatten()
    p = p.flatten()
    q = q.flatten()

    detector_pointing_f.mult_fortran_f(p, q, pq, n)

    pq = pq.reshape(shape)
    return pq


def arraylist_dot(a, b):
    """
    Dot product of ndarrays.
    p and q can have any size.
    Returns a column array.

    Parameters
    ----------
    a : array of arrays
        first list of arrays.
    b : array of arrays
        second list of arrays.

    Examples
    ----------
    >>> a = np.array([[1, 2, 3]])
    >>> b = np.array([[4, 6, 8], [3, 7, 1]])
    >>> arraylist_dot(a, b) #doctest: +NORMALIZE_WHITESPACE
    array([[40],
           [20]])
    """
    if a.ndim == 1 and b.ndim == 1:
        return np.array([[np.dot(a, b)]])
    else:
        return np.sum(a * b, axis=1)[:, np.newaxis]


def euler_quatx(alpha):
    """
    Generate quaternion units along x axis

    Parameters
    ----------
    alpha : float
        Polar angle in radian.

    Examples
    ----------
    >>> euler_quatx(alpha=np.pi/2.)
    array([ 0.70710678,  0.        ,  0.        ,  0.70710678])

    """
    alpha = np.asarray(alpha)
    z = np.zeros_like(alpha)
    c = np.cos(alpha * 0.5)
    s = np.sin(alpha * 0.5)
    return np.array([s, z, z, c]).T


def euler_quaty(alpha):
    """
    Generate quaternion units along y axis

    Parameters
    ----------
    alpha : float
        Polar angle in radian.

    Examples
    ----------
    >>> euler_quaty(alpha=np.pi/2.)
    array([ 0.        ,  0.70710678,  0.        ,  0.70710678])

    """
    alpha = np.asarray(alpha)
    z = np.zeros_like(alpha)
    c = np.cos(alpha * 0.5)
    s = np.sin(alpha * 0.5)
    return np.array([z, s, z, c]).T


def euler_quatz(alpha):
    """
    Generate quaternion units along z axis

    Parameters
    ----------
    alpha : float
        Polar angle in radian.

    Examples
    ----------
    >>> euler_quatz(alpha=np.pi/2.)
    array([ 0.        ,  0.        ,  0.70710678,  0.70710678])

    """
    alpha = np.asarray(alpha)
    z = np.zeros_like(alpha)
    c = np.cos(alpha * 0.5)
    s = np.sin(alpha * 0.5)
    return np.array([z, z, s, c]).T


def quat_to_radecpa_fortran(seq):
    """
    Routine to compute phi/theta/psi from a sequence
    of quaternions. Computation is done in fortran.

    WARNING: you still need to convert phi/theta/psi to get to RA/Dec/PA.

    Parameters
    ----------
    seq : array of arrays
        Array of quaternions.

    Returns
    ----------
    phi : 1d array
    theta : 1d array
    psi : 1d array
    """
    q1, q2, q3, q0 = seq.T
    n = q0.size
    phi = np.zeros_like(q0)
    theta = np.zeros_like(q0)
    psi = np.zeros_like(q0)

    detector_pointing_f.quat_to_radecpa_fortran_f(q0, q1, q2, q3, phi, theta, psi, n)
    return phi, theta, psi


def quat_to_radecpa_python(seq):
    """
    Routine to compute phi/theta/psi from a sequence
    of quaternions. Computation is done in Python.

    WARNING: you still need to convert phi/theta/psi to get to RA/Dec/PA.

    Parameters
    ----------
    seq : array of arrays
        Array of quaternions.

    Returns
    ----------
    phi : 1d array
    theta : 1d array
    psi : 1d array

    """
    q1, q2, q3, q0 = seq.T
    phi = np.arctan2(2 * (q0 * q1 + q2 * q3), 1.0 - 2.0 * (q1 * q1 + q2 * q2))
    theta = np.arcsin(2 * (q0 * q2 - q3 * q1))
    psi = np.arctan2(2 * (q0 * q3 + q1 * q2), 1.0 - 2.0 * (q2 * q2 + q3 * q3))
    return phi, theta, psi


def load_fake_pointing():
    """
    Load fake pointing parameters for testing purposes.

    Returns
    ----------
    allowed_params : list of string
        Name of the pointing model parameters used in `value_params`.
    value_params : 1d array
        Value of the pointing model parameters (see instrument.py).
        In degrees.
    az_enc : 1d array
        Encoder azimuth in radians.
    el_enc : 1d array
        Encoder elevation in radians.
    time : 1d array
        Encoder time (UTC) in mjd

    """
    allowed_params = "ia ie ca an aw"
    value_params = [10.28473073, 8.73953334, -15.59771781, -0.50977716, 0.10858016]
    az_enc = np.array([np.sin(2 * np.pi * i / 100) for i in range(100)])
    el_enc = np.ones(100) * 0.5
    time = np.array([56293 + t / 84000.0 for t in range(100)])

    return allowed_params, value_params, az_enc, el_enc, time


def update_ut1utc(fname="ut1utc_user.ephem", begin_date="2012-01-01", end_date=None):
    """
    Creates an updated database file with time correction to UTC.

    The script retrieves the IERS data base through astropy and requires access
    to the internet the first time this routine is called. Subsequent calls to
    IERS data base may not update the database.

    See https://docs.astropy.org/en/stable/utils/iers.html for more details.

    Parameters
    ----------
    fname : string
        Filename of the updated data base.
    begin_date : string
        First date to include in the updated data base. Supports any format
        supported by astropy Time class.
    end_date : string
        Last date to include in the updated data base. Supports any format
        supported by astropy Time class. By default include all the available
        dates in the IERS data base (usually ~1yr data in the future from the
        current date)
    """
    iers_table = iers.IERS_Auto.open()
    begin_time = Time(begin_date)
    mask = iers_table["MJD"].value >= begin_time.mjd
    if end_date is not None:
        end_time = Time(end_date)
        mask &= iers_table["MJD"].value <= end_time.mjd
    y = iers_table["year"][mask]
    m = iers_table["month"][mask]
    d = iers_table["day"][mask]
    mjd = iers_table["MJD"].value[mask]
    ut1utc = iers_table["UT1_UTC"][mask]

    # Corrects MJD according to astropy doc
    y[mjd <= 51543] += 1900
    y[mjd >= 51544] += 2000

    try:
        fname_path = os.path.join(os.environ["s4cmbPATH"], "s4cmb/data")
    except KeyError as e:
        print(str(e), "Saving the current directory instead")
        fname_path = "./"

    np.savetxt(
        os.path.join(fname_path, fname),
        np.column_stack((y, m, d, mjd, ut1utc)),
        fmt="%d-%d-%d\t%.6f\t%.6f",
    )
    iers_table.close()
    return


if __name__ == "__main__":
    import doctest

    if np.__version__ >= "1.14.0":
        np.set_printoptions(legacy="1.13")
    doctest.testmod()



#!/usr/bin/python
# Copyright (c) 2016-2021 Julien Peloton, Giulio Fabbian.
#
# This file is part of s4cmb
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Script to predict the trajectory of a body (Sun, Moon, ...) on the sky
[(theta, phi) coordinates] over the course of a year.

Author: Julien Peloton, peloton@lal.in2p3.fr
"""
from __future__ import division, absolute_import, print_function

import ephem
from datetime import datetime, date, time, timedelta
import numpy as np


class celestial_trajectory:
    """Predict the trajectory of a body (Sun, Moon, ...) """

    def __init__(
        self,
        body,
        lon_observer,
        lat_observer,
        elevation_observer,
        year,
        tz_offset=0.0,
    ):
        """
        Parameters
        ----------
        body : ephem.<body> instance
            <body> can be Sun(), Moon(), and so on. See ephem.
        lon_observer : str
            Longitute (angle) of the telescope. String form: 0:00:00.0.
        lat_observer : str
            Latitude (angle) of the telescope. String form: 0:00:00.0.
        elevation_observer : float
            Height above sea level (in meter).
        year : int
            Year of observation
        tz_offset : float
            Time correction to UTC. Default is 0.
        """
        self.body = body
        self.lon_observer = lon_observer
        self.lat_observer = lat_observer
        self.elevation_observer = elevation_observer
        self.year = year
        self.tz_offset = tz_offset

        self.months = [
            "Jan",
            "Feb",
            "Mar",
            "Apr",
            "May",
            "Jun",
            "Jul",
            "Aug",
            "Sep",
            "Oct",
            "Nov",
            "Dec",
        ]

        self.altaz = []
        self.radec = []
        self.thetaphi = []

        self.initialise_observer()
        self.traj_of_body_oneyear()

    def initialise_observer(self):
        """
        Set the Longitute and Latitude of the Observer.
        """
        # Define the observer
        self.ob = ephem.Observer()
        self.ob.lat, self.ob.lon = self.lat_observer, self.lon_observer
        self.ob.elevation = self.elevation_observer

    @staticmethod
    def radec2thetaphi(radec):
        """
        Correspondance between RA/Dec and theta/phi coordinate systems.

        Parameters
        ----------
        radec : tuple
            RA and Dec angles in radian.

        Returns
        ----------
        theta : float or 1d array
            Theta angle in radian
        phi : float or 1d array
            Phi angle in radian
        """
        theta = np.pi / 2 - radec[1]
        phi = float(radec[0])
        return (theta, phi)

    def radec_of(self, altaz):
        """
        Get RA/Dec from alt/az.

        Parameters
        ----------
        altaz : tuple
            Tuple containing (alt, az) in radians.

        Returns
        ----------
        RA : float
            RA of the body as seen by the observer [radian].
        Dec : float
            Dec of the body as seen by the observer [radian].
        """
        return self.ob.radec_of(altaz[1], altaz[0])

    def alt_az(self, date):
        """
        Compute altitude (elevation) and azimuth of the body as seen
        by the observer.

        Parameters
        ----------
        date : Python datetime instance
            Floating point value used by ephem to represent a date.
            The value is the number of days since 1899 December 31 12:00 UT.
            When creating an instance you can pass in a Python datetime
            instance, timetuple, year-month-day triple, or a plain float.
            Run str() on this object to see the UTC date it represents.
            ...
            WTF?

        Returns
        ----------
        alt : float
            Altitude (elevation) of the body at the time of observation as
            seen by the observer [radian].
        az : float
            Azimuth of the body at the time of observation as seen
            by the observer [radian].
        """
        self.ob.date = date
        self.body.compute(self.ob)
        return (self.body.alt, self.body.az)

    def traj_of_body_oneday(self, date):
        """
        Compute the coordinate (theta, phi) of the Sun at a particular date

        Parameters
        ----------
        date : Python datetime instance
            Floating point value used by ephem to represent a date.
            The value is the number of days since 1899 December 31 12:00 UT. When
            creating an instance you can pass in a Python datetime instance,
            timetuple, year-month-day triple, or a plain float.
            Run str() on this object to see the UTC date it represents.
            ...
            WTF?

        """
        # Update the date
        date = datetime.combine(date, time(12)) - timedelta(hours=self.tz_offset)
        self.ob.date = date

        # Alt/Az
        ALTAZ = self.alt_az(date)
        self.altaz.append(ALTAZ)

        # RA/Dec
        RADEC = self.radec_of(ALTAZ)
        self.radec.append(RADEC)

        # Theta/Phi
        THETAPHI = self.radec2thetaphi(RADEC)
        self.thetaphi.append(THETAPHI)

    def traj_of_body_oneyear(self):
        """
        Plot the course of the sun over one year.
        We assume each month is made of 28 days.

        Returns
        ----------
        coords : 2D array of floats
            The (theta, phi) coordinate for all days of the year. We compute
            the coordinate for only 28 days per month.
        months : list of strings
            Name of the 12 months.

        Examples
        ----------
        >>> sun = ephem.Sun()
        >>> sun_traj = celestial_trajectory(sun, '-67:46.816',
        ...     '-22:56.396', 5200, 2013)
        >>> print('RA =', round(sun_traj.thetaphi[0][0], 2))
        RA = 1.97

        """
        for month in range(1, 13):
            for day in range(1, 28):
                now = date(self.year, 1, 1).replace(month=month, day=day)
                self.traj_of_body_oneday(now)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



#!/usr/bin/python
# Copyright (c) 2016-2021 Julien Peloton, Giulio Fabbian.
#
# This file is part of s4cmb
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Module to handle instrument systematics.

Author: Julien Peloton, peloton@lal.in2p3.fr
        Giulio Fabbian, g.fabbian@sussex.ac.uk
"""
from __future__ import division, absolute_import, print_function

import numpy as np
from scipy import signal

from s4cmb.systematics_f import systematics_f
from s4cmb.instrument import construct_beammap, coordinates_on_grid, gauss2d

arcsecond2rad = np.pi / 180.0 / 3600.0
arcmin2rad = np.pi / 180.0 / 60.0
deg2rad = np.pi / 180.0


def inject_crosstalk_inside_SQUID(
    bolo_data,
    squid_ids,
    bolo_ids,
    mu=-3.0,
    sigma=1.0,
    radius=1,
    beta=2,
    seed=5438765,
    new_array=None,
    language="python",
):
    """
    Introduce leakage between neighboring bolometers within a SQUID.
    You have to provide the list of bolometers, to which SQUID they
    belong, and the index of bolometers within each SQUID.

    Parameters
    ----------
    bolo_data : list of arrays
        Contains the time-ordered data of bolometers.
    squid_ids : list of strings
        Contains the SQUID id for each bolometer.
        Should have the same shape[0] as bolo_data.
    bolo_ids : list of strings
        Contains the positions of bolometers within their SQUID.
        Should have the same shape[0] as bolo_data.
    mu : float, optional
        Mean of the Gaussian used to generate the level of leakage,
        in percent. E.g. mu=1.0 means the leakage coefficients will be
        centred around 1%.
    sigma : float, optional
        Width of the Gaussian used to generate the level of leakage,
        in percent. E.g. sigma=1.0 means the leakage coefficients will be
        of order X \pm 1%.
    radius : int, optional
        Controls the number of bolometers talking within a SQUID.
        radius=n means bolometers [N-n, N-n+1, ..., N, ..., N+n-1, N+n]
        will be crosstalking. `radius` must be at most equal to the number of
        bolometers within a SQUID.
    beta : int, optional
        Exponent controling the attenuation. The idea is that crosstalk depends
        on the separation length. Bolometer $b_i$ crosstalking with bolometers
        $b_{i \pm d}$ ($d \leq FF$) with leakage coefficients
        $\alpha_{i \pm d}$ will see his timestream modified as:
            $b_i = b_i + \dfrac{\alpha_{i \pm d}}{d^\beta}  b_{i \pm d}$.
        Assuming some fixed frequency spacing, hardware consideration
        suggests $\beta = 2$.
    seed : int, optional
        Control the random seed used to generate leakage coefficients.
    new_array : None or bolo_data-like array, optional
        If not None, return a new array of timestreams with the modifications.
        Modify bolo_data directly otherwise. Default is None.
    language : string, optional
        Language to perform computations to be chosen in ['python', 'fortran'].
        Default is python. [fortran is very slow...]

    Example
    ----------
    >>> from s4cmb.tod import load_fake_instrument
    >>> from s4cmb.tod import TimeOrderedDataPairDiff
    >>> inst, scan, sky_in = load_fake_instrument()
    >>> tod = TimeOrderedDataPairDiff(inst, scan, sky_in, CESnumber=0)
    >>> d = np.array([tod.map2tod(det) for det in range(2 * tod.npair)])
    >>> print(round(d[0][0], 3))
    40.95

    Inject crosstalk between neighbour bolometers (radius=1)
    >>> squid_ids = inst.focal_plane.get_indices('Sq')
    >>> bolo_ids = inst.focal_plane.bolo_index_in_squid
    >>> inject_crosstalk_inside_SQUID(d, squid_ids, bolo_ids, radius=1)
    >>> print(round(d[0][0], 3))
    39.561

    One can also keep the original timestreams, and return a new array
    containing modified timestreams.
    >>> d = np.array([tod.map2tod(det) for det in range(2 * tod.npair)])
    >>> d_new = np.zeros_like(d)
    >>> inject_crosstalk_inside_SQUID(d, squid_ids, bolo_ids, radius=1,
    ...     new_array=d_new, language='python')
    >>> print(round(d[0][0], 3), round(d_new[0][0], 3))
    ... #doctest: +NORMALIZE_WHITESPACE
    40.95 39.561

    For large number of bolometers per SQUID, you would prefer fortran
    to python to perform the loops. Choose python otherwise.
    """
    # Make mu and sigma unitless (user provides percentage)
    mu = mu / 100.0
    sigma = sigma / 100.0

    combs = {}
    for bolo in range(len(bolo_data)):
        sq = squid_ids[bolo]
        if sq not in combs:
            combs[sq] = []
        combs[sq].append((bolo_ids[bolo], bolo))

    tsout = 0.0 + bolo_data
    # How much to leak from one bolometer to its neighboring channels
    state = np.random.RandomState(seed)
    cross_amp = state.normal(mu, sigma, len(bolo_data))

    if language == "python":
        for sq in combs:
            for ch, i in combs[sq]:
                for ch2, i2 in combs[sq]:
                    separation_length = abs(ch - ch2)
                    if separation_length > 0 and separation_length <= radius:
                        tsout[i] += (
                            cross_amp[i2] / separation_length ** beta * tsout[i2]
                        )

    elif language == "fortran":
        # F2PY convention
        tsout = np.array(tsout, order="F")
        for sq in combs:
            local_indices = np.array(combs[sq]).flatten()[::2]
            global_indices = np.array(combs[sq]).flatten()[1::2]
            systematics_f.inject_crosstalk_inside_squid_f(
                tsout,
                local_indices,
                global_indices,
                radius,
                cross_amp,
                beta,
                len(local_indices),
                len(bolo_data),
                len(bolo_data[0]),
            )

    if new_array is not None:
        new_array[:] = tsout
    else:
        bolo_data[:] = tsout


def inject_crosstalk_SQUID_to_SQUID(
    bolo_data,
    squid_ids,
    bolo_ids,
    mu=-3.0,
    sigma=1.0,
    squid_attenuation=100.0,
    seed=5438765,
    new_array=None,
):
    """
    Introduce leakage between bolometers from different SQUIDs.
    You have to provide the list of bolometers, to which SQUID they
    belong, and the index of bolometers within each SQUID.

    Parameters
    ----------
    bolo_data : list of arrays
        Contains the time-ordered data of bolometers.
    squid_ids : list of strings
        Contains the SQUID id for each bolometer.
        Should have the same shape[0] as bolo_data.
    bolo_ids : list of strings
        Contains the positions of bolometers within their SQUID.
        Should have the same shape[0] as bolo_data.
    mu : float, optional
        Mean of the Gaussian used to generate the level of leakage,
        in percent. E.g. mu=1.0 means the leakage coefficients will be
        of order 1%.
    sigma : float, optional
        Width of the Gaussian used to generate the level of leakage,
        in percent. E.g. sigma=1.0 means the leakage coefficients will be
        of order X \pm 1%.
    squid_attenuation : int, optional
        Value controling the attenuation due to distance between SQUIDS.
        The idea is that crosstalk depends on the separation
        length between SQUIDs. Bolometer $b_i$ crosstalking with bolometers
        $b_{j}$ with leakage coefficients
        $\alpha_{j}$ will see his timestream modified as:
            $b_i = b_i + \dfrac{\alpha_{j}}{squid_attenuation}  b_{j}$.
        Default is 100.
    seed : int, optional
        Control the random seed used to generate leakage coefficients.
    new_array : None or bolo_data-like array, optional
        If not None, return a new array of timestreams with the modifications.
        Modify bolo_data directly otherwise. Default is None.

    Example
    ----------
    >>> from s4cmb.tod import load_fake_instrument
    >>> from s4cmb.tod import TimeOrderedDataPairDiff
    >>> inst, scan, sky_in = load_fake_instrument(nsquid_per_mux=4)
    >>> tod = TimeOrderedDataPairDiff(inst, scan, sky_in, CESnumber=0)
    >>> d = np.array([tod.map2tod(det) for det in range(2 * tod.npair)])
    >>> print(round(d[0][0], 3))
    40.948

    Inject crosstalk between bolometers in different SQUIDs
    >>> squid_ids = inst.focal_plane.get_indices('Sq')
    >>> bolo_ids = inst.focal_plane.bolo_index_in_squid
    >>> inject_crosstalk_SQUID_to_SQUID(d, squid_ids, bolo_ids,
    ...     squid_attenuation=100.)
    >>> print(round(d[0][0], 3))
    40.716

    One can also keep the original timestreams, and return a new array
    containing modified timestreams.
    >>> d = np.array([tod.map2tod(det) for det in range(2 * tod.npair)])
    >>> d_new = np.zeros_like(d)
    >>> inject_crosstalk_SQUID_to_SQUID(d, squid_ids, bolo_ids,
    ...     squid_attenuation=100., new_array=d_new)
    >>> print(round(d[0][0], 3), round(d_new[0][0], 3))
    ... #doctest: +NORMALIZE_WHITESPACE
    40.948 40.716

    """
    # Make mu and sigma unitless (user provides percentage)
    mu = mu / 100.0
    sigma = sigma / 100.0

    combs = {}
    for bolo in range(len(bolo_data)):
        sq = squid_ids[bolo]
        if sq not in combs:
            combs[sq] = []
        combs[sq].append((bolo_ids[bolo], bolo))

    tsout = 0.0 + bolo_data
    # How much to leak from one bolometer to its neighboring channels
    state = np.random.RandomState(seed)
    cross_amp = state.normal(mu, sigma, len(bolo_data))

    # Take one squid
    for sq1 in combs:
        # Take all bolometers in that SQUID
        for ch, i in combs[sq1]:
            # Take a second SQUID
            for sq2 in combs:
                if sq2 == sq1:
                    continue
                # Take all bolometers in that SQUID
                for ch2, i2 in combs[sq2]:
                    tsout[i] += cross_amp[i2] / squid_attenuation * tsout[i2]

    if new_array is not None:
        new_array[:] = tsout
    else:
        bolo_data[:] = tsout


def modify_beam_offsets(
    bolometer_xpos,
    bolometer_ypos,
    mu_diffpointing=10.0,
    sigma_diffpointing=5.0,
    seed=5847,
):
    """
    Modify the beam offsets (inject differential pointing between
    two pixel-pair bolometers). The model is the following:
    * Draw from a normal distribution G(mu, sigma)
      the magnitudes of the differential pointing rho.
    * Draw from a uniform distribution U(0, 2pi) the directions
      of the differential pointing theta.
    * Move the position of bottom bolometers as
      - x_top/bottom = \pm rho / 2 * cos(theta)
      - y_top/bottom = \pm rho / 2 * sin(theta)

    Parameters
    ----------
    bolometer_xpos : 1d array
        Initial beam centroid positions along the x axis. In radian.
    bolometer_ypos : 1d array
        Initial beam centroid positions along the y axis. In radian.
    mu_diffpointing : float, optional
        Mean of the distribution for differential pointing in arcsecond.
    sigma_diffpointing : float, optional
        Width of the distribution for differential pointing in arcsecond.
    seed : int, optional
        Random seed to control numbers generation.

    Returns
    ----------
    bolometer_xpos : 1d array
        Modified beam centroid positions along the x axis. In radian.
    bolometer_ypos : 1d array
        Modified beam centroid positions along the y axis. In radian.

    Examples
    ----------
    Inject differential pointing with magnitude 10 arcmin +- 5 arcmin.
    (completely irrealistic! just for test purposes).
    >>> bolometer_xpos = np.array([1., 1., -1., -1.]) * np.pi / 180.
    >>> bolometer_ypos = np.array([1., -1., -1., 1.]) * np.pi / 180.
    >>> x, y = modify_beam_offsets(bolometer_xpos, bolometer_ypos,
    ...     mu_diffpointing=600., sigma_diffpointing=300., seed=5847)
    >>> print(x * 180 / np.pi, y * 180 / np.pi) #doctest: +NORMALIZE_WHITESPACE
    [ 1.03802729  0.96197271 -1.02689298 -0.97310702]
    [ 0.92369044 -0.92369044 -1.10310547  1.10310547]

    """
    assert len(bolometer_xpos) == len(bolometer_ypos), ValueError(
        "x and y bolometer coordinates should have the same length"
    )
    state = np.random.RandomState(seed)

    npair = int(len(bolometer_xpos) / 2)

    # rho is defined by the user [xpos, ypos in radians - convert it!]
    rho = state.normal(
        mu_diffpointing * arcsecond2rad, sigma_diffpointing * arcsecond2rad, npair
    )

    # Angle are uniformly drawn from 0 and 360 degree
    theta = state.uniform(0, 360, npair) * deg2rad

    rhox = rho / 2.0 * np.cos(theta)
    rhoy = rho / 2.0 * np.sin(theta)

    bolometer_xpos[::2] += rhox
    bolometer_xpos[1::2] += -rhox
    bolometer_ypos[::2] += rhoy
    bolometer_ypos[1::2] += -rhoy

    return bolometer_xpos, bolometer_ypos


def inject_beam_ellipticity(
    sigma_gaussian,
    mu_beamellipticity,
    sigma_beamellipticity,
    nbolo,
    do_diffbeamellipticity=True,
    seed=54875,
):
    """
    Inject differential beam ellipticity (and beam ellipticity)
    Starting from the definition of the beam ellipticity:
        * eps = (sig1^2 - sig2^2)/(sig1^2 + sig2^2)
        * Assume ellipticity follows normal distribution
          centered on X% with Y% width
        * Assume sig1 = sig2 + d, where d is a small deviation
        * relate d and eps
    Furthermore, we assume the angle of the ellipses are taken
    from an uniform distribution btw -90 and 90 degrees.

    Parameters
    ----------
    sigma_gaussian : float
        Initial beam size (sigma). [radian]
    mu_beamellipticity : float
        Mean of the beam ellipticity distribution. [in percent]
        Between -100 and 100 percent.
    sigma_beamellipticity : float
        Width of the beam ellipticity distribution. [in percent]
    nbolo : int
        Number of bolometers for which beams have to be perturbed.
    do_diffbeamellipticity : bool, optional
        If False, bolometers in the same pair will have the same ellipticity.
        If True, add differential beam ellipticity between two bolometers.
        Default is True.
    seed : int, optional
        Seed for random number generation.

    Returns
    ----------
    sig_1 : 1d array of float
        Vector containing new values for semi-major axes for all bolometers
        in an elliptical Gaussian representation of the beam.
    sig_2 : 1d array of float
        Vector containing new values for semi-minor axes for all bolometers
        in an elliptical Gaussian representation of the beam.
    ellip_ang : 1d array of float
        Angles of rotation of beams around 0 for all bolometers.

    Examples
    ----------
    >>> from s4cmb.instrument import BeamModel
    >>> from s4cmb.instrument import FocalPlane
    >>> fp = FocalPlane(npair_per_squid=2, verbose=False)
    >>> bm = BeamModel(fp, verbose=False)

    Unperturbed beams (2 pairs of bolometers)
    >>> print(bm.sig_1)
    [ 0.00043235  0.00043235  0.00043235  0.00043235]

    Different pairs have different ellipticity
    (but not differential ellipticity within pairs)
    >>> sig_1, sig_2, ellip_ang = inject_beam_ellipticity(
    ...     bm.sig_1[0], 10, 5, 4, do_diffbeamellipticity=False)
    >>> print(sig_1)
    [ 0.0004715   0.0004715   0.00045276  0.00045276]

    Different pairs have different ellipticity and differential ellipticity
    within pairs
    >>> sig_1, sig_2, ellip_ang = inject_beam_ellipticity(
    ...     bm.sig_1[0], 10, 5, 4, do_diffbeamellipticity=True)
    >>> print(sig_1)
    [ 0.0004715   0.00045571  0.00045276  0.00046666]

    """
    state = np.random.RandomState(seed)

    eps = state.normal(
        mu_beamellipticity / 100.0, sigma_beamellipticity / 100.0, nbolo
    )

    if do_diffbeamellipticity:
        # d_plus = 2 * sigma_gaussian / eps * (1.0 + np.sqrt((1 - eps ** 2)))
        d_minus = 2 * sigma_gaussian / eps * (1.0 - np.sqrt((1 - eps ** 2)))
        d = d_minus
    else:
        # Bolometers in the same pair will have the same ellipticity.
        # This assumes that bolometers i and i+1 belong
        # to the same pair (default behaviour)
        eps = np.repeat(eps[::2], 2)
        # d_plus = 2 * sigma_gaussian / eps * (1.0 + np.sqrt((1 - eps ** 2)))
        d_minus = 2 * sigma_gaussian / eps * (1.0 - np.sqrt((1 - eps ** 2)))
        d = d_minus

    sig_1 = np.ones(nbolo) * sigma_gaussian + d / 2.0
    sig_2 = np.ones(nbolo) * sigma_gaussian - d / 2.0

    # Angle between ellipses
    ellip_ang = state.uniform(-90, 90, nbolo)

    return sig_1, sig_2, ellip_ang


def modify_pointing_parameters(values, errors):
    """
    This routine is not realistic at all for the moment!

    Parameters
    ----------
    values : 1d array
        Array containing values of the pointing parameters in degree.
    errors : 1d array
        Array containing values of the pointing parameter errors in degree.

    Returns
    ----------
    values_mod : 1d array
        Array containing modified values of the pointing parameters in degree.

    """
    values_mod = [p + err for p, err in zip(values, errors)]
    return values_mod


def step_function(nbolos, nsamples, mean=1, std=0.05, nbreaks=1, sign="same", seed=0):
    """
    Generate step functions for each bolometer from 1 to a values
    drawn from N(mean, std). The full timestream is broken into nbreaks
    corresponding to retuning of bolometers (gains are reset to one
    after a break). Return the gains for all bolometers (2 bolo in a
    pair will have the same break to avoid differential gain).

    Parameters
    ----------
    nbolos : int
        The number of bolometers
    nsamples : int
        Length of timestreams
    mean : float
        Mean of the distribution to sample drifts from. Default is 1.
    std : float
        Std of the distribution to sample drifts from.
        Default is 0.05 (that is \pm 5% with respect to a mean=1).
    nbreaks : int
        Number of break (number of retuning).
    sign : string
        If same, both detectors in the pair will have the same gain.
        If opposite, detectors will have gains with opposite sign
        (differential gain).

    Returns
    ----------
    gains : 2D array of size (nbolos, nsamples)
        The gains at each timestep for all bolometers

    Examples
    ----------
    >>> nbolos = 4
    >>> nsamples = 4
    >>> gains = step_function(nbolos, nsamples, nbreaks=1)
    >>> print(gains[0])
    [ 1.          1.          1.08820262  1.08820262]

    >>> gains = step_function(nbolos, nsamples, nbreaks=1, sign='opposite')
    >>> assert gains[0][0] == 2 - gains[1][0]
    """
    # Fix the seed
    state = np.random.RandomState(seed)

    # Initialise gains to ones
    gains = np.ones((nbolos, nsamples))

    # Length of each break
    length = nsamples // nbreaks
    sublength = nsamples // (2 * nbreaks)

    for pos in range(nbreaks):
        # 2 bolo in a pair will have the same break to avoid differential gain
        end_points = state.normal(mean, std, size=int(nbolos / 2))
        end_points = np.tile(end_points, (2, 1)).T.reshape((2 * len(end_points), 1))

        # Assign values
        shift = pos * length
        if pos * length < nsamples:
            gains[:, shift + sublength: shift + 2 * sublength] = end_points.reshape(
                (len(end_points), 1)
            )
        else:
            continue

        if sign == "opposite":
            gains[1] = 2 - gains[1]

    return gains


def step_function_gen(nsamples, mean=1, std=0.05, nbreaks=1, sign="same", seed=0):
    """
    Generator of step functions for each bolometer from 1 to a values
    drawn from N(mean, std). The full timestream is broken into nbreaks
    corresponding to retuning of bolometers (gains are reset to one
    after a break). Return the gains for 2 bolometers (2 bolo in a
    pair will have the same break to avoid differential gain).

    Particularly useful to save memory when simulating a lot of detectors.

    Parameters
    ----------
    nsamples : int
        Length of timestreams
    mean : float
        Mean of the distribution to sample drifts from. Default is 1.
    std : float
        Std of the distribution to sample drifts from.
        Default is 0.05 (that is \pm 5% with respect to a mean=1).
    nbreaks : int
        Number of break (number of retuning).
    sign : string
        If same, both detectors in the pair will have the same gain.
        If opposite, detectors will have gains with opposite sign
        (differential gain).

    Returns
    ----------
    gains : 2D array of size (2, nsamples)
        The gains at each timestep for 2 bolometers in a pair

    Examples
    ----------
    >>> nsamples = 4
    >>> gains_gen = step_function_gen(nsamples, nbreaks=1)
    >>> print(next(gains_gen)[0])
    [ 1.          1.          1.08820262  1.08820262]

    >>> gains_gen = step_function_gen(nsamples, nbreaks=1, sign='opposite')
    >>> g = next(gains_gen)
    >>> assert g[0][0] == 2 - g[1][0]
    """
    # Fix the seed
    state = np.random.RandomState(seed)

    # Initialise gains to ones
    gains = np.ones((2, nsamples))

    # Length of each break
    length = nsamples // nbreaks
    sublength = nsamples // (2 * nbreaks)

    while 1:
        for pos in range(nbreaks):
            # 2 bolo in a pair will have the same break
            # to avoid differential gain
            end_points = state.normal(mean, std, size=1)
            end_points = np.tile(end_points, (2, 1)).T.reshape((2, 1))

            # Assign values
            shift = pos * length
            if pos * length < nsamples:
                gains[
                    :, shift + sublength: shift + 2 * sublength
                ] = end_points.reshape((len(end_points), 1))
            else:
                continue

        if sign == "opposite":
            gains[1] = 2 - gains[1]

        yield gains


def linear_function(
    nbolos, nsamples, mean=1, std=0.05, nbreaks=1, sign="same", seed=0
):
    """
    Generate linear functions for each bolometer from 1 to a values
    drawn from N(mean, std). The full timestream is broken into nbreaks
    corresponding to retuning of bolometers (gains are reset to one
    after a break). Return the gains for all bolometers (2 bolo in a
    pair will have the same break to avoid differential gain)

    Parameters
    ----------
    nbolos : int
        The number of bolometers
    nsamples : int
        Length of timestreams
    mean : float
        Mean of the distribution to sample drifts from. Default is 1.
    std : float
        Std of the distribution to sample drifts from.
        Default is 0.05 (that is \pm 5% with respect to a mean=1).
    nbreaks : int
        Number of break (number of retuning).
    sign : string
        If same, both detectors in the pair will have the same gain.
        If opposite, detectors will have gains with opposite sign
        (differential gain).

    Returns
    ----------
    gains : 2D array of size (nbolos, nsamples)
        The gains at each timestep for all bolometers

    Examples
    ----------
    >>> nbolos = 4
    >>> nsamples = 4
    >>> gains = linear_function(nbolos, nsamples, nbreaks=1)
    >>> print(gains[0])
    [ 1.          1.02940087  1.05880174  1.08820262]

    >>> gains = linear_function(nbolos, nsamples, nbreaks=1, sign='opposite')
    >>> assert gains[0][0] == 2 - gains[1][0]
    """
    # Fix the seed
    state = np.random.RandomState(seed)

    # Initialise gains to ones
    gains = np.ones((nbolos, nsamples))

    # Length of each break
    length = nsamples // nbreaks

    for pos in range(nbreaks):
        # 2 bolo in a pair will have the same break to avoid differential gain
        end_points = state.normal(mean, std, size=int(nbolos / 2))
        end_points = np.tile(end_points, (2, 1)).T.reshape((2 * len(end_points), 1))

        # Assign values
        shift = pos * length
        if pos * length < nsamples:
            gains[:, shift: shift + length] = np.array(
                [
                    np.interp(
                        range(shift, shift + length),
                        [shift, shift + length - 1],
                        [1, end[0]],
                    )
                    for end in end_points
                ]
            )
        else:
            continue

    if sign == "opposite":
        gains[1] = 2 - gains[1]

    return gains


def linear_function_gen(nsamples, mean=1, std=0.05, nbreaks=1, sign="same", seed=0):
    """
    Generator returning linear functions for each bolometer from 1 to a values
    drawn from N(mean, std). The full timestream is broken into nbreaks
    corresponding to retuning of bolometers (gains are reset to one
    after a break). Yield the gains for 2 bolometers (2 bolo in a
    pair will have the same break to avoid differential gain).

    Particularly useful to save memory when simulating a lot of detectors.

    Parameters
    ----------
    nsamples : int
        Length of timestreams
    mean : float
        Mean of the distribution to sample drifts from. Default is 1.
    std : float
        Std of the distribution to sample drifts from.
        Default is 0.05 (that is \pm 5% with respect to a mean=1).
    nbreaks : int
        Number of break (number of retuning).
    sign : string
        If same, both detectors in the pair will have the same gain.
        If opposite, detectors will have gains with opposite sign
        (differential gain).

    Returns
    ----------
    gains : 2D array of size (2, nsamples)
        The gains at each timestep for 2 bolometers in a pair

    Examples
    ----------
    >>> nsamples = 4
    >>> gains_gen = linear_function_gen(nsamples, nbreaks=1)
    >>> print(next(gains_gen)[0])
    [ 1.          1.02940087  1.05880174  1.08820262]

    >>> gains_gen = linear_function_gen(nsamples, nbreaks=1, sign='opposite')
    >>> g = next(gains_gen)
    >>> assert g[0][0] == 2 - g[1][0]

    """
    # Fix the seed
    state = np.random.RandomState(seed)

    # Initialise gains to ones
    gains = np.ones((2, nsamples))

    # Length of each break
    length = nsamples // nbreaks

    while 1:
        for pos in range(nbreaks):
            # 2 bolo in a pair will have the same
            # break to avoid differential gain
            end_points = state.normal(mean, std, size=1)
            end_points = np.tile(end_points, (2, 1)).T.reshape((2, 1))

            # Assign values
            shift = pos * length
            if pos * length < nsamples:
                gains[:, shift: shift + length] = np.array(
                    [
                        np.interp(
                            range(shift, shift + length),
                            [shift, shift + length - 1],
                            [1, end[0]],
                        )
                        for end in end_points
                    ]
                )
            else:
                continue

        if sign == "opposite":
            gains[1] = 2 - gains[1]

        yield gains


def get_kernel_coefficients(
    beamprm, pairlist, kernel_type="diff", nx=128, pix_size=None, basis="circular"
):
    """
    Generate beam expansion coefficients.

    Here we compute the coefficients of the operator K to go from the observed
    temperature map to the T->P and P->T as well as T->T, P->P leakage. It is
    based on the sum and difference beam maps derivative expansion onto
    a basis beam:

    diffbeam = K- conv basis_beam,
    sumbeam  = K+ conv basis_beam,

    with

    K+/- = a.D,

    where D is a vector containing beam map and its derivatives:

    D = (I, dI/dt, dI/dp, d2I/dpdt, d2I/dt2, d2I/d2p),

    and a are coefficients. This assumes the polarized beams are the same for
    intensity and linear polarization.

    Parameters
    ----------
    beamprm : beam_model instance
        Instance of beam_model.
    pairlist : list of list
        List containing the indices of bolometers grouped by pair
        [[0, 1], [2, 3], ...]
    kernel_type : string
        Specify if the kernel needed ar for a sum or diff beams or both.
        Supported options: 'sum','diff' or both.
    nx : int, optional
        Number of pixels per row/column (in pixel) to construct the beam maps.
        You want nx * pix_size to be a large number compared to the beam size
        to make sure to incorporate all beam features.
    pix_size : float, optional
        Pixel size in radian. If None, set automatically to be 1/7th of the
        beam size.
    basis : string
        Specify the basis onto whish developing the sum or diff beam. The
        default assumes a Gaussian circular beam with FWHM specified in the beam
        model instance. The sum beam can be used instead.
        Supported options: 'circular' (default),'sumbeam'

    Returns
    ----------
    out : array of array
        Array containing beam kernel coefficients
        (one for each T derivative) for all pairs.

    Examples
    ----------
    >>> from s4cmb.instrument import FocalPlane
    >>> from s4cmb.instrument import BeamModel
    >>> fp = FocalPlane(verbose=False)
    >>> pairlist = np.reshape(fp.bolo_index_in_fp, (fp.npair, 2))

    If the beams are identical, the coefficients will be zeros
    >>> bm = BeamModel(fp, verbose=False)
    >>> coeffs = get_kernel_coefficients(bm, pairlist, nx=32)
    >>> np.allclose(coeffs[0], np.zeros_like(coeffs[0]))
    True

    Let's perturb the beam now by adding beam ellipticity.
    Beam ellipticity involves a Stokes parameter and its second derivatives,
    but not the first derivatives (2nd and 3rd coefficients are zeros).
    >>> bm = BeamModel(fp, verbose=False)
    >>> bm.sig_1, bm.sig_2, bm.ellip_ang = inject_beam_ellipticity(
    ...     bm.sig_1[0], 10, 5, fp.nbolometer, do_diffbeamellipticity=True)
    >>> coeffs = get_kernel_coefficients(bm, pairlist, nx=32)

    Let's perturb the beam now by adding differential pointing.
    Differential pointing mainly the first derivatives of a Stokes parameter,
    but not the temperature itself and its second derivatives
    (0th, 4th, 5th, and 6th coefficients are zeros).
    >>> bm = BeamModel(fp, verbose=False)
    >>> bm.xpos, bm.ypos = modify_beam_offsets(bm.xpos, bm.ypos,
    ...     mu_diffpointing=600., sigma_diffpointing=300., seed=5847)
    >>> coeffs = get_kernel_coefficients(bm, pairlist, nx=32)

    """
    assert basis in ["circular", "sumbeam"]
    assert kernel_type in ["sum", "diff", "sumdiff", "diffsum"]
    if pix_size is None:
        # Go from sigma to FWHM
        mean_FWHM_x = np.mean(beamprm.sig_1) / np.sqrt(8 * np.log(2))
        mean_FWHM_y = np.mean(beamprm.sig_2) / np.sqrt(8 * np.log(2))

        # 1/7th of the beam size.
        pix_size = (mean_FWHM_x + mean_FWHM_y) / 2.0 / 7.0

    if basis == "circular":
        # Creates the a circular beam used to develop the sum/diff beam
        fwhm = beamprm.fwhm * np.pi / 180 / 60
        xy2f = coordinates_on_grid(pix_size=pix_size, nx=nx)
        basis_beam = gauss2d(
            xy2f,
            0,
            0,
            1,
            fwhm / np.sqrt(8 * np.log(2)),
            fwhm / np.sqrt(8 * np.log(2)),
            0,
        ).reshape((nx, nx))

    out_diff = []
    out_sum = []

    for ct, cb in pairlist:
        summap, diffmap = construct_beammap(beamprm, ct, cb, nx, pix_size)
        if summap is not None and diffmap is not None:
            if basis != "circular":
                basis_beam = summap
            if "diff" in kernel_type:
                kernel_diff = split_deriv(basis_beam, diffmap, pix_size)
            if "sum" in kernel_type:
                kernel_sum = split_deriv(basis_beam, summap, pix_size)
        else:
            kernel_sum = None
        if "diff" in kernel_type:
            out_diff.append(kernel_diff)
        if "sum" in kernel_type:
            out_sum.append(kernel_sum)
    if ("diff" in kernel_type) and ("sum" in kernel_type):
        return np.array(out_diff), np.array(out_sum)
    else:
        if "diff" == kernel_type:
            return np.array(out_diff)
        if "sum" == kernel_type:
            return np.array(out_sum)


def split_deriv(sumbeam, diffbeam, pix_size):
    """
    Try to split a convolution kernel diffbeam (B-) in to two convolutions,
    one low order derivatives (K), and the other a base kernel (B+):

    B- = K conv B+

    where K = a.D, with D is a vector containing Stokes parameter map
    and its derivatives. In the case of e.g. temperature:

    D = (I, dI/dt, dI/dp, d2I/dpdt, d2I/dt2, d2I/d2p),

    and a are coefficients.

    Parameters
    ----------
    sumbeam : 2d array
        Sum beam map
    diffbeam : 2d array
        Difference beam map
    pix_size : float
        Size of a pixel [radian]

    Returns
    ----------
    x : 1d array of 6 elements
        Vector containing the beam kernel coefficients a
        (one for each derivative of the beam map).

    Examples
    ----------
    >>> from s4cmb.instrument import FocalPlane
    >>> from s4cmb.instrument import BeamModel
    >>> fp = FocalPlane(verbose=False)
    >>> bm = BeamModel(fp, verbose=False)
    >>> pix_size = 0.5 / 60. * np.pi / 180.
    >>> summap, diffmap = construct_beammap(bm, 0, 1, nx=32, pix_size=pix_size)
    >>> K = split_deriv(summap, diffmap, pix_size)

    """
    ds = derivs(sumbeam, pix_size)

    A = ds.reshape((6, sumbeam.shape[0] * sumbeam.shape[1])).T
    b = diffbeam.reshape(diffbeam.shape[0] * diffbeam.shape[1])

    x, residues, rank, s = np.linalg.lstsq(A, b, rcond=-1)

    return x


def xderiv(m, pix_size):
    """
    Perform the derivative of a 2d map with respect to x coordinate.

    Parameters
    ----------
    m : 2d array
        Input 2d map
    pix_size : float
        Size of a pixel [radian]

    Returns
    ----------
    -v : 2d array
        Derivative of the input beam map with respect to x coordinate.

    Examples
    ----------
    >>> m = np.ones((10, 10)) * np.arange(10)
    >>> dmx = xderiv(m, pix_size=0.5/60.*np.pi/180.)
    """
    kernel = np.array([[1, 0, -1]]) / (2.0 * pix_size)
    v = signal.convolve2d(m, kernel, mode="same")
    return -v


def yderiv(m, pix_size):
    """
    Perform the derivative of a 2d map with respect to y coordinate.

    Parameters
    ----------
    m : 2d array
        Input 2d map
    pix_size : float
        Size of a pixel [radian]

    Returns
    ----------
    -v : 2d array
        Derivative of the input beam map with respect to y coordinate.

    Examples
    ----------
    >>> m = np.ones((10, 10)) * np.arange(10)
    >>> dmy = yderiv(m.T, pix_size=0.5/60.*np.pi/180.)
    """
    kernel = np.array([[1, 0, -1]]).T / (2.0 * pix_size)
    v = signal.convolve2d(m, kernel, mode="same")
    return -v


def derivs(m, pix_size):
    """
    Compute full 1st and 2nd derivatives of a 2d map (flat sky).

    Parameters
    ----------
    m : 2d array
        Input 2d map
    pix_size : float
        Size of a pixel [radian]

    Returns
    ----------
    v : array of 2d array
        Derivatives of the input beam map with respect to x and y coordinates.

    Examples
    ----------
    >>> m = np.ones((10, 10)) * np.arange(10)
    >>> m = m.T * np.arange(10)
    >>> m00, m10, m01, m11, m20, m02 = derivs(m, pix_size=0.5/60.*np.pi/180.)
    """
    m00 = m
    m10 = xderiv(m, pix_size)
    m01 = yderiv(m, pix_size)
    m11 = yderiv(m10, pix_size)
    m20 = xderiv(m10, pix_size)
    m02 = yderiv(m01, pix_size)

    return np.array((m00, m10, m01, m11, m20, m02))


def fixspin(K, spin):
    """
    Selects derivatives of only a given spin (by default we do not compute
    all spins if this is unecessary) and rearrange for later use.

    Parameters
    ----------
    K : 1d array
        Kernel coefficients of the beam map decomposition
    spins : string
        Which type of derivatives you want to use for you leakage.
        Example: beam ellipticity would use 0 and 2, while differential
        pointing would use just 1.

    Returns
    ----------
    es : 1d array
        Selected kernel coefficients of the beam map decomposition

    Examples
    ----------
    >>> from s4cmb.instrument import FocalPlane
    >>> from s4cmb.instrument import BeamModel
    >>> fp = FocalPlane(verbose=False)
    >>> bm = BeamModel(fp, verbose=False)
    >>> pix_size = 0.5 / 60. * np.pi / 180.
    >>> summap, diffmap = construct_beammap(bm, 0, 1, nx=32, pix_size=pix_size)
    >>> K = split_deriv(summap, diffmap, pix_size)
    >>> SK = fixspin(K, spin='012')

    """
    d00, d10, d01, d11, d20, d02 = K
    x = (d20 + d02) * 0.5
    y = (d20 - d02) * 0.5

    e00 = 0.0
    e10 = 0.0
    e01 = 0.0
    e11 = 0.0
    e20 = 0.0
    e02 = 0.0

    if "0" in spin:
        e00 += d00
        e10 += 0.0
        e01 += 0.0
        e11 += 0.0
        e20 += x
        e02 += x
    if "1" in spin:
        e00 += 0.0
        e10 += d10
        e01 += d01
        e11 += 0.0
        e20 += 0.0
        e02 += 0.0
    if "2" in spin:
        e00 += 0.0
        e10 += 0.0
        e01 += 0.0
        e11 += d11
        e20 += y
        e02 += -y

    return np.array((e00, e10, e01, e11, e20, e02))


def rotate_deriv(K, theta):
    """
    Rotate a vector of derivative kernel coefficients K by an angle theta.

    Parameters
    ----------
    K : 1d array
        Kernel coefficients of the beam map decomposition
    theta : float
        Orientation of the pixel [radian]

    Returns
    ----------
    es : 1d array
        Rotated kernel coefficients of the beam map decomposition.

    Examples
    ----------
    >>> from s4cmb.instrument import FocalPlane
    >>> from s4cmb.instrument import BeamModel
    >>> fp = FocalPlane(verbose=False)
    >>> bm = BeamModel(fp, verbose=False)
    >>> pix_size = 0.5 / 60. * np.pi / 180.
    >>> summap, diffmap = construct_beammap(bm, 0, 1, nx=32, pix_size=pix_size)
    >>> K = split_deriv(summap, diffmap, pix_size)
    >>> RK = rotate_deriv(K, theta=np.pi)

    """

    d00, d10, d01, d11, d20, d02 = K

    c = np.cos(theta)
    s = np.sin(theta)

    e00 = np.ones_like(theta) * d00

    e10 = c * d10 - s * d01
    e01 = s * d10 + c * d01

    e20 = c * c * d20 - 2.0 * c * s * d11 + s * s * d02
    e02 = s * s * d20 + 2.0 * c * s * d11 + c * c * d02
    e11 = (c * c - s * s) * d11 + c * s * (d20 - d02)

    es = np.array((e00, e10, e01, e11, e20, e02))

    return es


def fix_kernel_type(K, kernel_type="diffnomonopole"):
    if "sum" in kernel_type:
        # subtracts monopole part already included
        # in the input timestream
        K[:, 0] -= 1
    if "nomonopole" in kernel_type:
        # Force monopole to 0 when using leakage beam
        K[:, 0] = 0.0
    # Flips the sign of dtheta terms.
    K[:, 1] *= -1
    K[:, 3] *= -1
    return K


def waferts_add_diffbeam(
    waferts,
    point_matrix,
    beam_orientation,
    intensity_derivatives,
    diffbeam_kernels,
    pairlist,
    spins="012",
    kernel_type="diff_nomonopole",
    pol_derivatives=None,
    pol_angle=None,
    spins_pol="012",
    Usign=1,
):
    """
    Modify timestreams by injecting T->P leakage from beam mismatch.
    Note that timestreams are modified on-the-fly.

    Parameters
    ----------
    waferts : ndarray
        Array containing timestreams
    point_matrix : ndarray
        Array containing pointing information
    beam_orientation : ndarray
        Array containing intrinsic orientation of the pixels
    intensity_derivatives : ndarray
        Containing map of the intensity and its derivatives (6 maps). They need
        to include only the pixels in the field defined by the scan (see obspix
        field of TOD class).
    diffbeam_kernels : 1d array of 6 elements
        Array containing beam kernel coefficients
        (one for each T derivative) for all pairs.
    pairlist : list of list
        List containing the indices of bolometers grouped by pair
        [[0, 1], [2, 3], ...].
    spins : string
        Which type of derivatives you want to use for you leakage.
        Example: beam ellipticity would use 0 and 2, while differential
        pointing would use just 1.
    kernel_type : string
        String specifying if sum or diff beam leakage components
        has to be injected. Add _nomonopole if you want not to include
        monopole leakage component and force it to be 0.
    pol_derivatives : list of ndarray
        Containing map of the Q and U parameters and their
        derivatives (2x6 maps: [[Q,dQ/dx...],[U,dU/dx...]])
    pol_angle: ndarray
        Array containing the value of the polarization angle
    spins_pol: string
        as the spins argument. Allows to use different leakage
        derivatives for the polarization part of the beam leakage. By
        default spins=spins_pol
    Usign: int
        defines sign convention to match the IAU convention in case
        the flat map projection is used.

    Examples
    ----------
    >>> from s4cmb.tod import load_fake_instrument
    >>> from s4cmb.tod import TimeOrderedDataPairDiff
    >>> inst, scan, sky_in = load_fake_instrument(compute_derivatives=True)

    Make beam ellipticity
    >>> sig_1, sig_2, ellip_ang = inject_beam_ellipticity(
    ...     inst.beam_model.sig_1[0], 50, 10,
    ...     inst.focal_plane.nbolometer, do_diffbeamellipticity=True)
    >>> inst.beam_model.sig_1 = sig_1
    >>> inst.beam_model.sig_2 = sig_2
    >>> inst.beam_model.ellip_ang = ellip_ang

    Get TOD
    >>> tod = TimeOrderedDataPairDiff(inst, scan, sky_in, CESnumber=0)
    >>> d = np.array([tod.map2tod(det) for det in range(2 * tod.npair)])

    Get beam kernels
    >>> pairlist = np.reshape(
    ...     inst.focal_plane.bolo_index_in_fp, (inst.focal_plane.npair, 2))
    >>> K = get_kernel_coefficients(
    ...     inst.beam_model, pairlist, nx=128, pix_size=None)

    Get intensity derivatives and orientation of pixels
    >>> intensity_derivatives = np.array(
    ...     [sky_in.I, sky_in.dIdt, sky_in.dIdp,
    ...      sky_in.d2Idpdt, sky_in.d2Id2t, sky_in.d2Id2p])
    >>> beam_orientation = np.array(
    ...     [tod.pol_angs[ch] - (
    ...         90.0 - tod.intrinsic_polangle[2*ch]) * np.pi / 180. -
    ...      2*tod.hwpangle for ch in range(inst.focal_plane.npair)])

    Inject spurious signal
    >>> waferts_add_diffbeam(
    ...     d, tod.point_matrix, beam_orientation,
    ...     intensity_derivatives, K,
    ...     pairlist, spins='012')

    """
    assert kernel_type in ["sum", "diff", "sum_nomonopole", "diff_nomonopole"]

    # Just to preserve it
    K = diffbeam_kernels + 0.0

    K = fix_kernel_type(K, kernel_type)

    if "sum" in kernel_type:
        # adds I->I in sum timestream and P->P in diff timestream
        bottom_sign = -1
    else:
        bottom_sign = 1

    for i in range(len(K)):
        K[i] = fixspin(K[i], spins)

    diffbeamleak = np.zeros((int(waferts.shape[0] / 2), waferts.shape[1]))

    # adds I leakage. if kernel_type is "diff" adds I->P leakage, otherwise
    # adds I->I leakage
    diffbeam_map2tod(
        diffbeamleak, intensity_derivatives, point_matrix, beam_orientation, K
    )

    waferts[::2] += diffbeamleak
    waferts[1::2] -= bottom_sign * diffbeamleak

    # adds P leakage. If kernel_type is "diff" adds P->I leakage, otherwise
    # adds P->P leakage
    if (pol_derivatives is not None) and (pol_angle is not None):
        if (spins_pol != spins) and (pol_derivatives is not None):
            # initialize a new copy of polarization kernels withot modifying
            # the input
            K_pol = fix_kernel_type(diffbeam_kernels + 0.0, kernel_type)
            K_pol[i] = fixspin(K_pol[i], spins_pol)
        else:
            K_pol = K
        diffbeamleak[:, :] = 0.0
        diffbeam_map2tod(
            diffbeamleak,
            pol_derivatives,
            point_matrix,
            beam_orientation,
            K_pol,
            pol_angle,
            Usign,
        )

        waferts[::2] += diffbeamleak
        waferts[1::2] += bottom_sign * diffbeamleak


def diffbeam_map2tod(
    out,
    signal_derivatives,
    point_matrix,
    beam_orientation,
    diffbeam_kernels,
    pol_angle=None,
    Usign=1,
):
    """
    Scan the leakage maps to generate polarisation timestream for channel ch,
    using the differential beam model kernel.


    Note for future:
     * check flat sky!

    Parameters
    ----------
    out : ndarray
        Will contain the spurious signal (npair, nsamples)
    signal_derivatives : ndarray
        Containing map of the intensity and its derivatives (6 maps)
        or a list of Q and U derivatives (2x6 maps)
    point_matrix : ndarray
        Array containing pointing information
    beam_orientation : ndarray
        Array containing intrinsic orientation of the pixels
    diffbeam_kernels : 1d array of 6 elements
        Array containing beam kernel coefficients
        (one for each T derivative) for all pairs.
    pol_angle: ndarray
        Array containing polarization angle information.
    Usign: int
        defines sign convention to match the IAU convention in case
        the flat map projection is used.
    """
    npix, nt = out.shape
    assert point_matrix.shape == out.shape
    assert diffbeam_kernels.shape[0] == npix
    if pol_angle is None:
        assert len(signal_derivatives) == diffbeam_kernels.shape[1]
    else:
        assert len(signal_derivatives) == 2
        assert len(signal_derivatives[0]) == diffbeam_kernels.shape[1]
    assert beam_orientation.shape == (npix, nt)

    for ipix in range(npix):
        i1d = point_matrix[ipix, :]
        okpointing = i1d != -1
        io = i1d[okpointing]
        k = rotate_deriv(diffbeam_kernels[ipix], -beam_orientation[ipix, okpointing])
        if pol_angle is not None:
            # compute cosine and sine polarization angles to save time
            psi = pol_angle[ipix, :]
            c2pol_ang = np.cos(2 * psi[okpointing])
            s2pol_ang = np.sin(2 * psi[okpointing])
            if signal_derivatives is not None:
                for coeff in range(len(k)):
                    # Q derivatives
                    der0 = signal_derivatives[0][coeff, io]
                    out[ipix, okpointing] += der0 * k[coeff] * c2pol_ang
                    # U derivatives
                    der1 = signal_derivatives[1][coeff, io]
                    out[ipix, okpointing] += der1 * k[coeff] * s2pol_ang * Usign
        # Add the leakage on-the-fly
        else:
            for coeff in range(len(k)):
                # T derivatives
                out[ipix, okpointing] += signal_derivatives[coeff, io] * k[coeff]


if __name__ == "__main__":
    import doctest

    if np.__version__ >= "1.14.0":
        np.set_printoptions(legacy="1.13")
    doctest.testmod()



#!/usr/bin/python
# Copyright (c) 2016-2021 Julien Peloton, Giulio Fabbian.
#
# This file is part of s4cmb
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Script to generate normalized inputs for the software x2pure
(pure pseudo-spectrum estimator).

Author: Julien Peloton, peloton@lal.in2p3.fr
"""
from __future__ import division, absolute_import, print_function

import os
import numpy as np
import errno

from s4cmb.input_sky import write_healpix_cmbmap


def safe_mkdir(path, verbose=False):
    """
    Create a path and catch the race condition between path exists and mkdir.

    Parameters
    ----------
    path : string
        Name of the folder to be created.
    verbose : bool
        If True, print messages about the status.

    Examples
    ----------
    Folders are created
    >>> safe_mkdir('toto/titi/tutu', verbose=True)

    Folders aren't created because they already exist
    >>> safe_mkdir('toto/titi/tutu', verbose=True)
    Folders toto/titi/tutu already exist. Not created.

    >>> os.removedirs('toto/titi/tutu')
    """
    abspath = os.path.abspath(path)
    if not os.path.exists(abspath):
        try:
            os.makedirs(abspath)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise
    else:
        if verbose:
            print("Folders {} already exist. Not created.".format(path))


def qu_weight_mineig(cc, cs, ss, epsilon=0.0, verbose=False):
    """
    Create a weight map using the smallest eigenvalue
    of the polarization matrix.

    Parameters
    ----------
    cc : 1d array
        Projected noise weighted cosine**2 per-pixel.
    ss : 1d array
        Projected noise weighted sine**2 per-pixel.
    cs : 1d array
        Projected noise weighted cosine * sine per-pixel.
    epsilon : float, optional
        Threshold for selecting the pixels. 0 <= epsilon < 1/4.
        The higher the more selective.

    Returns
    ----------
    weights : 1d array
        Polarisation weights per-pixel.

    """
    trace = cc + ss
    trace2 = trace * trace
    det = cc * ss - cs * cs

    val2 = trace2 - 4 * det
    valid = val2 > 0.0
    val = np.zeros_like(val2)
    val[valid] = np.sqrt(val2[valid])

    weight = np.zeros_like(trace)
    lambda_minus = (trace - val) / 2

    if verbose:
        print("criterion is", epsilon, "< det < 1/4 (epsilon= 0. by default)")

    valid = lambda_minus > (trace - np.sqrt(trace2 - 4 * epsilon * trace2)) / 2

    if verbose:
        valid3 = [x for x in valid if x is True]
        print("number of pixels kept:", len(valid3), "/", np.sum(trace > 0))
        print(
            "Percentage cut: {:3f} %%".format(
                (1.0 - float(len(valid3)) / np.sum(trace > 0)) * 100.0
            )
        )

    weight[valid] = lambda_minus[valid]
    return weight


def write_maps_a_la_xpure(OutputSkyMap, name_out, output_path):
    """
    Grab sky data from OutputSkyMap and write them into files readable by
    the software xpure.

    Parameters
    ----------
    OutputSkyMap : OutputSkyMap instance
        Instance of OutputSkyMap containing map data.
    name_out : string
        File name (.fits) where to store the data.
    output_path : string
        Folder where to put the data.

    """
    safe_mkdir(os.path.join(output_path, name_out))

    fits_I = np.zeros((12 * OutputSkyMap.nside * OutputSkyMap.nside))
    fits_Q = np.zeros((12 * OutputSkyMap.nside * OutputSkyMap.nside))
    fits_U = np.zeros((12 * OutputSkyMap.nside * OutputSkyMap.nside))

    I, Q, U = OutputSkyMap.get_IQU()
    fits_I[OutputSkyMap.obspix] = I
    fits_Q[OutputSkyMap.obspix] = Q
    fits_U[OutputSkyMap.obspix] = U

    full_path = os.path.join(output_path, name_out, "IQU_{}.fits".format(name_out))

    write_healpix_cmbmap(
        full_path,
        data=[fits_I, fits_Q, fits_U],
        fits_IDL=False,
        coord=None,
        colnames=["I", "Q", "U"],
        partial=False,
        nest=False,
    )

    del fits_I, fits_Q, fits_U


def write_weights_a_la_xpure(OutputSkyMap, name_out, output_path, epsilon, HWP=False):
    """
    Grab weight and mask from OutputSkyMap and write them into files
    readable by the software xpure.

    Parameters
    ----------
    OutputSkyMap : OutputSkyMap instance
        Instance of OutputSkyMap containing map data.
    name_out : string
        File name (.fits) where to store the data.
    output_path : string
        Folder where to put the data.
    epsilon : float
        Threshold for selecting the pixels. 0 <= epsilon < 1/4.
        The higher the more selective.
    HWP : bool
        If True, use demodulation syntax for the weights (w0, w4).
        Default is False (pair difference syntax: w, cc, ss, cs)

    """
    safe_mkdir(os.path.join(output_path, name_out))

    # Intensity
    weight = np.zeros((12 * OutputSkyMap.nside * OutputSkyMap.nside))

    if not HWP:
        weight[OutputSkyMap.obspix] = OutputSkyMap.w
    else:
        weight[OutputSkyMap.obspix] = OutputSkyMap.w0

    full_path = os.path.join(output_path, name_out, "Iw_{}.fits".format(name_out))
    write_healpix_cmbmap(
        full_path,
        data=weight,
        fits_IDL=False,
        coord=None,
        colnames=["I"],
        partial=False,
        nest=False,
    )

    binary = np.zeros((12 * OutputSkyMap.nside * OutputSkyMap.nside))
    mask = np.where((weight > 0))[0]
    binary[mask] = 1.0
    full_path = os.path.join(
        output_path, name_out, "Iw_{}_norm.fits".format(name_out)
    )
    write_healpix_cmbmap(
        full_path,
        data=binary,
        fits_IDL=False,
        coord=None,
        colnames=["I"],
        partial=False,
        nest=False,
    )

    # Polarisation
    weight = np.zeros((12 * OutputSkyMap.nside * OutputSkyMap.nside))
    if not HWP:
        weight[OutputSkyMap.obspix] = qu_weight_mineig(
            OutputSkyMap.cc, OutputSkyMap.cs, OutputSkyMap.ss, epsilon
        )
    else:
        weight[OutputSkyMap.obspix] = OutputSkyMap.w4

    full_path = os.path.join(output_path, name_out, "Pw_{}.fits".format(name_out))
    write_healpix_cmbmap(
        full_path,
        data=weight,
        fits_IDL=False,
        coord=None,
        colnames=["P"],
        partial=False,
        nest=False,
    )

    binary = np.zeros((12 * OutputSkyMap.nside * OutputSkyMap.nside))
    mask = np.where((weight > 0))[0]
    binary[mask] = 1.0
    full_path = os.path.join(
        output_path, name_out, "Pw_{}_norm.fits".format(name_out)
    )
    write_healpix_cmbmap(
        full_path,
        data=binary,
        fits_IDL=False,
        coord=None,
        colnames=["P"],
        partial=False,
        nest=False,
    )


def create_batch(batch_file, name_out, params_s4cmb, params_xpure):
    """
    Create submission file for the software xpure

    Parameters
    ----------
    batch_file : string
        Filename.
    params_s4cmb : NormaliseS4cmbParser instance
        Object with s4cmb parameter values.
    params_xpure : NormaliseXpureParser instance
        Object with xpure parameter values.
    """
    host = os.environ["NERSC_HOST"]
    if host == "edison":
        params_xpure.nproc_per_node = 24
    elif host == "cori":
        params_xpure.nproc_per_node = 32

    with open(batch_file, "w") as f:
        print("#!/bin/bash -l", file=f)
        print("#SBATCH -p {}".format(params_xpure.queue), file=f)
        print("#SBATCH -N {}".format(params_xpure.node), file=f)
        print("#SBATCH -t {}".format(params_xpure.time), file=f)
        print("#SBATCH -J {}".format(name_out), file=f)
        if host == "cori":
            print("#SBATCH -C haswell", file=f)

        print(" ", file=f)
        print("#### TO BE CHANGED BY USER ####", file=f)
        print("# Name for files and folders", file=f)
        print("name={}".format(name_out), file=f)

        print(" ", file=f)
        print("# Radius for apodization", file=f)
        print("radius={}".format(params_xpure.radius_apodization), file=f)

        print(" ", file=f)
        print("# Parameters for XPURE", file=f)
        print("LMAX_USER={}".format(params_xpure.lmax_user), file=f)
        print("NSIDE_USER={}".format(params_s4cmb.nside_out), file=f)

        print(" ", file=f)
        print("# Number of simulation (put 1 if not simulation)", file=f)
        print("NSIMU_USER=1", file=f)

        print(" ", file=f)
        print(
            "# Number of masks (put one if several maps but one common masks)",
            file=f,
        )
        print("NMASK_USER=1", file=f)

        print(" ", file=f)
        print("# Number of maps (if > 1, will do the cross correlations)", file=f)
        print("NMAPS_USER=1", file=f)

        print(" ", file=f)
        print("# Beam and bins", file=f)
        print("BEAMFILE={}".format(params_xpure.beam_file), file=f)
        print("BINTAB={}".format(params_xpure.bin_file), file=f)

        print(" ", file=f)
        print("# XPURE running mode (0=xpol, 1=xpure, 2=hybrid)", file=f)
        print("MODE_XPURE={}".format(params_xpure.xpure_mode), file=f)

        print(" ", file=f)
        txt = "# FULL (0=myapodizemask, create_mll and XPURE) or FAST "\
            + "(1=only XPURE) or SEMI-FAST (=2 create_mll and XPURE)"
        print(
            txt,
            file=f,
        )
        print("FAST={}".format(params_xpure.fast), file=f)

        print(" ", file=f)
        print("#### END TO BE CHANGED BY USER ####", file=f)

        print(" ", file=f)
        print("#ENVIRONEMENT", file=f)
        print("SCRATCH2=/global/cscratch1/sd/peloton", file=f)
        if host == "edison":
            print("BINDIR=${HOME}/mapmaker/xpure/xpure/trunk/build/edison", file=f)
        elif host == "cori":
            print("BINDIR=${HOME}/mapmaker/xpure/xpure/trunk/build/cori", file=f)
        print("THEORYDIR=${SCRATCH2}/s4cmb/additional_files", file=f)
        print("OUTPUTMLL=${SCRATCH2}/s4cmb/xpure/cls/${name}", file=f)
        print("OUTPUTCL=${SCRATCH2}/s4cmb/xpure/cls/${name}", file=f)
        print("mkdir -p ${OUTPUTMLL}", file=f)
        print("mkdir -p ${OUTPUTCL}", file=f)
        print("MASKDIR=${SCRATCH2}/s4cmb/xpure/masks/${name}", file=f)
        print("MAPDIR=${SCRATCH2}/s4cmb/xpure/maps/${name}", file=f)

        print(" ", file=f)
        print("##########################", file=f)
        print("# MASK (COMMON MASK)", file=f)
        print("##########################", file=f)
        print("cd $MASKDIR", file=f)

        print(" ", file=f)
        print('binary_name=$(basename $(find -name "Iw*${name}_norm.fits"))', file=f)
        print('weight_name=$(basename $(find -name "Iw*${name}.fits"))', file=f)
        print("BINARY_MASK_I1=${MASKDIR}/${binary_name}", file=f)
        print("WEIGHT_I1=${MASKDIR}/${weight_name}", file=f)
        print("APODIZED_MASK_I1=${MASKDIR}/ap30_${weight_name}", file=f)

        print(" ", file=f)
        print('binary_name=$(basename $(find -name "Pw*${name}_norm.fits"))', file=f)
        print('weight_name=$(basename $(find -name "Pw*${name}.fits"))', file=f)
        print("BINARY_MASK_P1=${MASKDIR}/${binary_name}", file=f)
        print("WEIGHT_P1=${MASKDIR}/${weight_name}", file=f)
        print("APODIZED_MASK_P1=${MASKDIR}/ap30_${weight_name}", file=f)

        print(" ", file=f)
        print("## Intensity", file=f)
        print("output_spin0_I1=${MASKDIR}/spin0_I_${name}.fits", file=f)
        print("output_spin1_I1=${MASKDIR}/spin1_I_${name}.fits", file=f)
        print("output_spin2_I1=${MASKDIR}/spin2_I_${name}.fits", file=f)

        print(" ", file=f)
        print("## Polarization", file=f)
        print("output_spin0_P1=${MASKDIR}/spin0_P_${name}.fits", file=f)
        print("output_spin1_P1=${MASKDIR}/spin1_P_${name}.fits", file=f)
        print("output_spin2_P1=${MASKDIR}/spin2_P_${name}.fits", file=f)

        print(" ", file=f)
        print("##########################", file=f)
        print("# MAPS", file=f)
        print("##########################", file=f)
        print("cd ${MAPDIR}", file=f)

        print(" ", file=f)
        print("for (( i=1; i<=${NMAPS_USER}; i++)); do", file=f)
        print('	MAPS[${i}]=$(basename $(find -name "IQU*${name}.fits"))', file=f)
        print("done", file=f)

        print(" ", file=f)
        print("# CHECK", file=f)
        print("for (( i=1; i<=${NMAPS_USER}; i++)); do", file=f)
        print("	echo ${MAPS[${i}]}", file=f)
        print("done", file=f)

        print(" ", file=f)
        print("cd $PBS_O_WORKDIR", file=f)

        print(" ", file=f)
        print(
            "################################",
            file=f,
        )
        print("# From binary mask to binary apodized mask", file=f)
        print(
            "################################",
            file=f,
        )
        print('if [ "${FAST}" -eq "0" ]', file=f)
        print("	then", file=f)
        txt = "	time srun -N 1 -n {} ${{BINDIR}}/myapodizemask "\
            + "${{BINARY_MASK_I1}} ${{APODIZED_MASK_I1}} -minpix 1 -inside 1 "\
            + "-radius ${{radius}} & time srun -N 1 -n {} "\
            + "${{BINDIR}}/myapodizemask ${{BINARY_MASK_P1}} "\
            + "${{APODIZED_MASK_P1}} -minpix 1 -inside 1 -radius ${{radius}}"
        print(
            txt.format(
                params_xpure.nproc_apo, params_xpure.nproc_apo
            ),
            file=f,
        )
        print("	wait", file=f)
        print("else", file=f)
        print('	echo "Go fast - skip myapodizemask"', file=f)
        print("fi", file=f)

        print(" ", file=f)
        print(
            "################################",
            file=f,
        )
        print("# Computation of spin window function", file=f)
        print(
            "################################",
            file=f,
        )

        print(" ", file=f)
        print("cat > param_all_I11${name}.par << EOF", file=f)
        print("#healpix parameters", file=f)
        print("##########", file=f)
        print("nside = ${NSIDE_USER}", file=f)
        print("lmax = ${LMAX_USER} #should not exceed 2*Nside", file=f)

        print(" ", file=f)
        print("#input file parameters", file=f)
        print("############", file=f)
        print("maskBinary = ${BINARY_MASK_I1}", file=f)
        print("window_spin0 = ${APODIZED_MASK_I1}", file=f)
        print("inverseNoise = ${WEIGHT_I1}", file=f)

        print(" ", file=f)
        print("#output file parameters", file=f)
        print("############", file=f)
        print("output_spin0 = ${output_spin0_I1}", file=f)
        print("output_spin1 = ${output_spin1_I1}", file=f)
        print("output_spin2 = ${output_spin2_I1}", file=f)
        print("EOF", file=f)

        print(" ", file=f)
        print("cat > param_all_P11${name}.par << EOF", file=f)
        print("#healpix parameters", file=f)
        print("##########", file=f)
        print("nside = ${NSIDE_USER}", file=f)
        print("lmax = ${LMAX_USER} #should not exceed 2*Nside", file=f)

        print(" ", file=f)
        print("#input file parameters", file=f)
        print("############", file=f)
        print("maskBinary = ${BINARY_MASK_P1}", file=f)
        print("window_spin0 = ${APODIZED_MASK_P1}", file=f)
        print("inverseNoise = ${WEIGHT_P1}", file=f)

        print(" ", file=f)
        print("#output file parameters", file=f)
        print("############", file=f)
        print("output_spin0 = ${output_spin0_P1}", file=f)
        print("output_spin1 = ${output_spin1_P1}", file=f)
        print("output_spin2 = ${output_spin2_P1}", file=f)
        print("EOF", file=f)

        print(" ", file=f)
        print('if [ "${FAST}" -eq "0" ]', file=f)
        print("	then", file=f)
        txt = "	time srun -N {} -n {} ${{BINDIR}}/scalar2spin "\
            + "param_all_I11${{name}}.par >& output_scalar2spinI11${{name}} "\
            + "& time srun -N {} -n {} "\
            + "${{BINDIR}}/scalar2spin param_all_P11${{name}}.par "\
            + ">& output_scalar2spinP11${{name}}"
        print(
            txt.format(
                int(params_xpure.nproc_scalar_to_spin // params_xpure.nproc_per_node),
                params_xpure.nproc_scalar_to_spin,
                int(params_xpure.nproc_scalar_to_spin // params_xpure.nproc_per_node),
                params_xpure.nproc_scalar_to_spin,
            ),
            file=f,
        )
        print("	wait", file=f)
        print("else", file=f)
        print('	echo "Go fast - skip scalar2spin"', file=f)
        print("fi", file=f)

        print(" ", file=f)
        print(
            "################################",
            file=f,
        )
        print("# X2pure", file=f)
        print(
            "################################",
            file=f,
        )
        print("cd ${OUTPUTMLL}", file=f)

        print(" ", file=f)
        print(
            "################################",
            file=f,
        )
        print("# Mode-mode mixing matrix", file=f)
        print(
            "################################",
            file=f,
        )
        print("cat > createMll.par << EOF", file=f)

        print(" ", file=f)
        print("##### MODE #######", file=f)
        print("# 0 : Standard formalism", file=f)
        print("# 1 : Pure formalism", file=f)
        print("# 2 : Hybrid formalism", file=f)
        print("##############", file=f)
        print("mode = ${MODE_XPURE}", file=f)

        print(" ", file=f)
        print("###### SETUP #####", file=f)
        print("nside = ${NSIDE_USER}", file=f)
        print("lmax = ${LMAX_USER}", file=f)
        print("nmask = ${NMASK_USER}", file=f)

        print(" ", file=f)
        print("EOF", file=f)

        print(" ", file=f)
        print("for (( i=1; i<=${NMASK_USER}; i++)); do", file=f)
        print("	ind=$(($i - 1))", file=f)
        print("	cat >> createMll.par << EOF", file=f)

        print(" ", file=f)
        print("maskfile${i}_T  = ${output_spin0_I1}", file=f)

        print(" ", file=f)
        print("maskfile${i}_E_spin0 = ${output_spin0_P1}", file=f)
        print("maskfile${i}_E_spin1 = ${output_spin1_P1}", file=f)
        print("maskfile${i}_E_spin2 = ${output_spin2_P1}", file=f)

        print(" ", file=f)
        print("maskfile${i}_B_spin0 = ${output_spin0_P1}", file=f)
        print("maskfile${i}_B_spin1 = ${output_spin1_P1}", file=f)
        print("maskfile${i}_B_spin2 = ${output_spin2_P1}", file=f)

        print(" ", file=f)
        print("mllfile_TT_TT_${i} = ${OUTPUTMLL}/mll_TT_TT_BinMask${i}.fits", file=f)

        print(" ", file=f)
        print("mllfile_EE_EE_${i} = ${OUTPUTMLL}/mll_spinEE_EE_pcg${i}.fits", file=f)
        print("mllfile_EE_BB_${i} = ${OUTPUTMLL}/mll_spinEE_BB_pcg${i}.fits", file=f)
        print("mllfile_EE_EB_${i} = ${OUTPUTMLL}/mll_spinEE_EB_pcg${i}.fits", file=f)
        print("mllfile_BB_BB_${i} = ${OUTPUTMLL}/mll_spinBB_BB_pcg${i}.fits", file=f)
        print("mllfile_BB_EE_${i} = ${OUTPUTMLL}/mll_spinBB_EE_pcg${i}.fits", file=f)
        print("mllfile_BB_EB_${i} = ${OUTPUTMLL}/mll_spinBB_EB_pcg${i}.fits", file=f)

        print(" ", file=f)
        print("mllfile_TE_TE_${i} = ${OUTPUTMLL}/mll_spinTE_TE_pcg${i}.fits", file=f)
        print("mllfile_TE_TB_${i} = ${OUTPUTMLL}/mll_spinTE_TB_pcg${i}.fits", file=f)
        print("mllfile_TB_TE_${i} = ${OUTPUTMLL}/mll_spinTB_TE_pcg${i}.fits", file=f)
        print("mllfile_TB_TB_${i} = ${OUTPUTMLL}/mll_spinTB_TB_pcg${i}.fits", file=f)

        print(" ", file=f)
        print("mllfile_EB_EB_${i} = ${OUTPUTMLL}/mll_spinEB_EB_pcg${i}.fits", file=f)
        print("mllfile_EB_EE_${i} = ${OUTPUTMLL}/mll_spinEB_EE_pcg${i}.fits", file=f)
        print("mllfile_EB_BB_${i} = ${OUTPUTMLL}/mll_spinEB_BB_pcg${i}.fits", file=f)

        print(" ", file=f)
        print("EOF", file=f)

        print(" ", file=f)
        print("done", file=f)

        print(" ", file=f)
        print('if [ "${FAST}" -eq "0" ]', file=f)
        print("        then", file=f)
        txt = "	time srun -N {} -n {} "\
            + "${{BINDIR}}/x2pure_create_mll createMll.par"
        print(
            txt.format(
                int(params_xpure.nproc_mll // params_xpure.nproc_per_node),
                params_xpure.nproc_mll,
            ),
            file=f,
        )

        print("	rm -f createMll.par", file=f)
        print('elif [ "${FAST}" -eq "2" ]', file=f)
        print("        then", file=f)
        txt = "        time srun -N {} -n {} "\
            + "${{BINDIR}}/x2pure_create_mll createMll.par"
        print(
            txt.format(
                int(params_xpure.nproc_mll // params_xpure.nproc_per_node),
                params_xpure.nproc_mll,
            ),
            file=f,
        )
        print("        rm -f createMll.par", file=f)
        print("else", file=f)
        print('	echo "Go fast - skip x2pure_create_mll"', file=f)
        print("fi", file=f)

        print(" ", file=f)
        print(
            "#################################",
            file=f,
        )
        print("# X2pure", file=f)
        print(
            "################################",
            file=f,
        )
        print("for (( n=0; n<${NSIMU_USER}; n++ )); do", file=f)

        print(" ", file=f)
        print(
            '    echo "************************ simu $n ************************"',
            file=f,
        )

        print(" ", file=f)
        print("    num=$n", file=f)

        print(" ", file=f)
        print("    #CREATE PARAMETER FILE", file=f)
        print("    cat > xpure.par << _EOF_", file=f)

        print(" ", file=f)
        print("mode = ${MODE_XPURE}", file=f)

        print(" ", file=f)
        print("nside = ${NSIDE_USER}", file=f)
        print("nmaps = ${NMAPS_USER}", file=f)
        print("nmasks = ${NMASK_USER}", file=f)

        print(" ", file=f)
        print("_EOF_", file=f)

        print(" ", file=f)
        print("    for ((j=1; j<=${NMAPS_USER}; j++)); do", file=f)

        print(" ", file=f)
        print("	cat >> xpure.par << _EOF_", file=f)

        print(" ", file=f)
        print("bellfile${j} = ${THEORYDIR}/${BEAMFILE}", file=f)
        print("mapfile${j} = ${MAPDIR}/${MAPS[${j}]}", file=f)

        print(" ", file=f)
        print("_EOF_", file=f)

        print(" ", file=f)
        print("    done", file=f)

        print(" ", file=f)
        print("    cat >> xpure.par << _EOF_", file=f)

        print(" ", file=f)
        print("lmaxSim = ${LMAX_USER}", file=f)

        print(" ", file=f)
        print("_EOF_", file=f)

        print(" ", file=f)
        print("    for(( i=1; i<=${NMASK_USER}; i++)); do", file=f)
        print("	ind=$(($i - 1))", file=f)
        print("	cat >> xpure.par << EOF", file=f)

        print(" ", file=f)
        print("maskfile${i}_T  = ${output_spin0_I1}", file=f)

        print(" ", file=f)
        print("maskfile${i}_E_spin0 = ${output_spin0_P1}", file=f)
        print("maskfile${i}_E_spin1 = ${output_spin1_P1}", file=f)
        print("maskfile${i}_E_spin2 = ${output_spin2_P1}", file=f)

        print(" ", file=f)
        print("maskfile${i}_B_spin0 = ${output_spin0_P1}", file=f)
        print("maskfile${i}_B_spin1 = ${output_spin1_P1}", file=f)
        print("maskfile${i}_B_spin2 = ${output_spin2_P1}", file=f)

        print(" ", file=f)
        print("mllfile_TT_TT_${i} = ${OUTPUTMLL}/mll_TT_TT_BinMask${i}.fits", file=f)

        print(" ", file=f)
        print("mllfile_EE_EE_${i} = ${OUTPUTMLL}/mll_spinEE_EE_pcg${i}.fits", file=f)
        print("mllfile_EE_BB_${i} = ${OUTPUTMLL}/mll_spinEE_BB_pcg${i}.fits", file=f)
        print("mllfile_EE_EB_${i} = ${OUTPUTMLL}/mll_spinEE_EB_pcg${i}.fits", file=f)
        print("mllfile_BB_BB_${i} = ${OUTPUTMLL}/mll_spinBB_BB_pcg${i}.fits", file=f)
        print("mllfile_BB_EE_${i} = ${OUTPUTMLL}/mll_spinBB_EE_pcg${i}.fits", file=f)
        print("mllfile_BB_EB_${i} = ${OUTPUTMLL}/mll_spinBB_EB_pcg${i}.fits", file=f)

        print(" ", file=f)
        print("mllfile_TE_TE_${i} = ${OUTPUTMLL}/mll_spinTE_TE_pcg${i}.fits", file=f)
        print("mllfile_TE_TB_${i} = ${OUTPUTMLL}/mll_spinTE_TB_pcg${i}.fits", file=f)
        print("mllfile_TB_TE_${i} = ${OUTPUTMLL}/mll_spinTB_TE_pcg${i}.fits", file=f)
        print("mllfile_TB_TB_${i} = ${OUTPUTMLL}/mll_spinTB_TB_pcg${i}.fits", file=f)

        print(" ", file=f)
        print("mllfile_EB_EB_${i} = ${OUTPUTMLL}/mll_spinEB_EB_pcg${i}.fits", file=f)
        print("mllfile_EB_EE_${i} = ${OUTPUTMLL}/mll_spinEB_EE_pcg${i}.fits", file=f)
        print("mllfile_EB_BB_${i} = ${OUTPUTMLL}/mll_spinEB_BB_pcg${i}.fits", file=f)

        print(" ", file=f)
        print("EOF", file=f)
        print("    done", file=f)

        print(" ", file=f)
        print("    cat >> xpure.par << _EOF_", file=f)
        print("noise_biasT_1 = 0.", file=f)
        print("noise_biasT_2_2 = 0.", file=f)
        print("noise_biasT_1_2 = 0.", file=f)

        print(" ", file=f)
        print("noise_biasP_1 = 0.", file=f)
        print("noise_biasP_2_2 = 0.", file=f)
        print("noise_biasP_1_2 = 0.", file=f)

        print(" ", file=f)
        print("bintab = ${THEORYDIR}/${BINTAB}", file=f)
        print("#mask_list = ${INPUT}/bin2mask_43bins.fits", file=f)
        print("_EOF_", file=f)

        print(" ", file=f)
        print("cat >> xpure.par << _EOF_", file=f)
        print("pseudofile = ${OUTPUTCL}/pseudopure_pcg_${name}_$num", file=f)
        print("cellfile = ${OUTPUTCL}/cellpure_pbear_${name}_$num", file=f)
        print("lmax = ${LMAX_USER}", file=f)
        print("_EOF_", file=f)
        print("    #RUN", file=f)
        print(
            "    time srun -N {} -n {} ${{BINDIR}}/x2pure xpure.par".format(
                int(params_xpure.nproc_xpure // params_xpure.nproc_per_node),
                params_xpure.nproc_xpure,
            ),
            file=f,
        )

        print(" ", file=f)
        print("done", file=f)


if __name__ == "__main__":
    import doctest

    if np.__version__ >= "1.14.0":
        np.set_printoptions(legacy="1.13")
    doctest.testmod()



# -*- coding: utf-8 -*-
# Copyright 2016-2021 Julien Peloton, Giulio Fabbian
# Licensed under the GPL-3.0 License, see LICENSE file for details.
from s4cmb import __version__
from setuptools import find_packages
from numpy.distutils.core import setup
from numpy.distutils.misc_util import Configuration

def configuration(parent_package='', top_path=None):
    config = Configuration('s4cmb', parent_package, top_path)
    config.add_extension(
        'scanning_strategy_f',
        sources=['s4cmb/scanning_strategy_f.f90'],
        libraries=[], f2py_options=[],
        extra_f90_compile_args=[
            '-ffixed-line-length-1000',
            '-O3'
        ],
        extra_compile_args=[''],
        extra_link_args=['']
    )
    config.add_extension(
        'detector_pointing_f',
        sources=['s4cmb/detector_pointing_f.f90'],
        libraries=[], f2py_options=[],
        extra_f90_compile_args=[
            '-ffixed-line-length-1000',
            '-O3'
        ],
        extra_compile_args=[''],
        extra_link_args=[''],
    )
    config.add_extension(
        'tod_f',
        sources=['s4cmb/tod_f.f90'],
        libraries=[], f2py_options=[],
        extra_f90_compile_args=[
            '-ffixed-line-length-1000',
            '-O3'
        ],
        extra_compile_args=[''],
        extra_link_args=[''],
    )
    config.add_extension(
        'systematics_f',
        sources=['s4cmb/systematics_f.f90'],
        libraries=[], f2py_options=[],
        extra_f90_compile_args=[
            '-ffixed-line-length-1000',
            '-O3'],
        extra_compile_args=[''],
        extra_link_args=[''],
    )
    return config


if __name__ == "__main__":
    # Download url
    d_url = 'https://github.com/JulienPeloton/s4cmb/archive/{}.tar.gz'.format(
        __version__)

    reqs = open('requirements.txt', 'r').read().strip().splitlines()

    package_data = {
        's4cmb/data': [
            'test_data_set_lensedCls.dat',
            'ut1utc.ephem'
        ],
        'examples': [
            'simple_app.py',
            'so_app.py',
            'so_crosstalk_app.py',
            'so_gain_variation.py',
            'simple_parameters.ini',
            'so_parameters.ini',
            'xpure_parameters.ini',
            'so_flat_MC_app.py',
            'nersc_cori.batch'
        ]
    }

    setup(
        configuration=configuration,
        version=__version__,
        url='https://github.com/JulienPeloton/s4cmb',
        download_url=d_url,
        license='GPL-3.0',
        author='Julien Peloton, Giulio Fabbian',
        author_email='peloton@lal.in2p3.fr',
        description='Simulate systematic effects in the context of CMB',
        platforms='any',
        packages=find_packages(),
        include_package_data=True,
        package_data=package_data,
        install_requires=reqs,
        classifiers=[
            'Environment :: Console',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3',
            'Topic :: Scientific/Engineering :: Astronomy'
        ]
    )



# Copyright (c) 2016-2021 Julien Peloton, Giulio Fabbian.
#
# This file is part of s4cmb
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Module including diverse tools to manipulate alms and maps .
"""
import numpy as np
import healpy as hp


def get_healpix_ring_pixel_layout(nside, th_idx):
    """Healpix ring layout.

    From 'get_pixel_layout' subroutine in healpix f90 package.

    Author: Julien Carron (j.carron@sussex.ac.uk)
    """
    ith = th_idx + 1
    nrings = 2 * nside
    assert 1 <= ith <= 4 * nside - 1, (ith, nrings)
    if ith > nrings:
        startpix, nphi, kphi0, cth, sth = get_healpix_ring_pixel_layout(
            nside, ith - 2 * (ith - nrings) - 1
        )
        return 12 * nside ** 2 - startpix - nphi, nphi, kphi0, -cth, sth
    dth1 = 1.0 / 3.0 / nside ** 2
    dth2 = 2.0 / 3.0 / nside
    dst1 = 1.0 / (np.sqrt(6.0) * nside)
    if ith < nside:  # polar cap (north)
        cth = 1.0 - ith ** 2 * dth1
        nphi = 4 * ith
        kphi0 = 1
        sth = np.sin(2.0 * np.arcsin(ith * dst1))
        startpix = 2 * ith * (ith - 1)
    else:
        cth = (2 * nside - ith) * dth2
        nphi = 4 * nside
        kphi0 = (ith + 1 - nside) % 2
        sth = np.sqrt((1.0 - cth) * (1.0 + cth))
        startpix = 2 * nside * (nside - 1) + (ith - nside) * int(nphi)
    return startpix, nphi, kphi0, cth, sth


def get_alpha_raise(s, lmax):
    """Response coefficient of spin-s spherical harmonic to spin raising operator.

    Author: Julien Carron (j.carron@sussex.ac.uk)
    """
    ret = np.zeros(lmax + 1, dtype=float)
    ret[abs(s):] = np.sqrt(
        np.arange(abs(s) - s, lmax - s + 1) * np.arange(abs(s) + s + 1, lmax + s + 2)
    )
    return ret


def get_alpha_lower(s, lmax):
    """Response coefficient of spin-s spherical harmonic to spin lowering operator.

    Author: Julien Carron (j.carron@sussex.ac.uk)
    """
    ret = np.zeros(lmax + 1, dtype=float)
    ret[abs(s):] = -np.sqrt(
        np.arange(s + abs(s), lmax + s + 1) * np.arange(abs(s) - s + 1, lmax - s + 2)
    )
    return ret


def alm2map_spin_der1(gclm, nside, spin, zbounds=(-1.0, 1.0), ret_slice=None):
    """Returns spin-s transform '_{s}d' of alm,
    together with d/dtheta _{s}d and 1/sin tht d/dphi _{s}d.

    This crude version has three calls to spin-weight harmonics alm2map_spin.

    Author: Julien Carron (j.carron@sussex.ac.uk)
    """
    assert spin > 0, spin
    assert hp.Alm.getlmax(gclm[0].size) == hp.Alm.getlmax(gclm[1].size)
    lmax = hp.Alm.getlmax(gclm[0].size)
    zbounds = np.sort(np.array(zbounds))
    # shape (2, 12 * nside ** 2),
    # first entry = real part, second entry imaginary part.
    _sd = np.array(hp.alm2map_spin(gclm, nside, spin, lmax))
    if spin > 1:
        _gclm = [
            hp.almxfl(gclm[0], get_alpha_lower(spin, lmax)),
            hp.almxfl(gclm[1], get_alpha_lower(spin, lmax)),
        ]
        _sm1d = np.array(hp.alm2map_spin(_gclm, nside, spin - 1, lmax))
    else:
        _sm1d = -np.array(
            [
                hp.alm2map(
                    hp.almxfl(gclm[0], get_alpha_lower(spin, lmax)),
                    nside,
                    verbose=False,
                ),
                hp.alm2map(
                    hp.almxfl(gclm[1], get_alpha_lower(spin, lmax)),
                    nside,
                    verbose=False,
                ),
            ]
        )

    _gclm = [
        hp.almxfl(gclm[0], get_alpha_raise(spin, lmax)),
        hp.almxfl(gclm[1], get_alpha_raise(spin, lmax)),
    ]
    _sp1d = np.array(hp.alm2map_spin(_gclm, nside, spin + 1, lmax))

    d_dth = -0.5 * (_sp1d + _sm1d)

    d_dphi_sin0 = 0.5 * np.array([-_sp1d[1] + _sm1d[1], _sp1d[0] - _sm1d[0]])
    for iring in range(4 * nside - 1):
        startpix, nphi, kphi0, cth, sth = get_healpix_ring_pixel_layout(nside, iring)
        if zbounds[0] <= cth <= zbounds[1]:
            slic = slice(startpix, startpix + nphi)
            d_dphi_sin0[1, slic] -= spin * (cth / sth) * _sd[0, slic]
            d_dphi_sin0[0, slic] += spin * (cth / sth) * _sd[1, slic]
    if ret_slice is not None:
        return _sd[:, ret_slice], d_dth[:, ret_slice], d_dphi_sin0[:, ret_slice]

    return _sd, d_dth, d_dphi_sin0
